#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import nova.conf
from nova.notifications.objects import base
from nova.notifications.objects import flavor as flavor_payload
from nova.notifications.objects import keypair as keypair_payload
from nova.objects import base as nova_base
from nova.objects import fields


CONF = nova.conf.CONF


@nova_base.NovaObjectRegistry.register_notification
class InstancePayload(base.NotificationPayloadBase):
    SCHEMA = {
        'uuid': ('instance', 'uuid'),
        'user_id': ('instance', 'user_id'),
        'tenant_id': ('instance', 'project_id'),
        'reservation_id': ('instance', 'reservation_id'),
        'display_name': ('instance', 'display_name'),
        'display_description': ('instance', 'display_description'),
        'host_name': ('instance', 'hostname'),
        'host': ('instance', 'host'),
        'node': ('instance', 'node'),
        'os_type': ('instance', 'os_type'),
        'architecture': ('instance', 'architecture'),
        'availability_zone': ('instance', 'availability_zone'),

        'image_uuid': ('instance', 'image_ref'),

        'key_name': ('instance', 'key_name'),

        'kernel_id': ('instance', 'kernel_id'),
        'ramdisk_id': ('instance', 'ramdisk_id'),

        'created_at': ('instance', 'created_at'),
        'launched_at': ('instance', 'launched_at'),
        'terminated_at': ('instance', 'terminated_at'),
        'deleted_at': ('instance', 'deleted_at'),
        'updated_at': ('instance', 'updated_at'),

        'state': ('instance', 'vm_state'),
        'power_state': ('instance', 'power_state'),
        'task_state': ('instance', 'task_state'),
        'progress': ('instance', 'progress'),

        'metadata': ('instance', 'metadata'),
        'locked': ('instance', 'locked'),
        'auto_disk_config': ('instance', 'auto_disk_config')
    }
    # Version 1.0: Initial version
    # Version 1.1: add locked and display_description field
    # Version 1.2: Add auto_disk_config field
    # Version 1.3: Add key_name field
    # Version 1.4: Add BDM related data
    # Version 1.5: Add updated_at field
    # Version 1.6: Add request_id field
    # Version 1.7: Added action_initiator_user and action_initiator_project to
    #              InstancePayload
    # Version 1.8: Added locked_reason field
    VERSION = '1.8'
    fields = {
        'uuid': fields.UUIDField(),
        'user_id': fields.StringField(nullable=True),
        'tenant_id': fields.StringField(nullable=True),
        'reservation_id': fields.StringField(nullable=True),
        'display_name': fields.StringField(nullable=True),
        'display_description': fields.StringField(nullable=True),
        'host_name': fields.StringField(nullable=True),
        'host': fields.StringField(nullable=True),
        'node': fields.StringField(nullable=True),
        'os_type': fields.StringField(nullable=True),
        'architecture': fields.StringField(nullable=True),
        'availability_zone': fields.StringField(nullable=True),

        'flavor': fields.ObjectField('FlavorPayload'),
        'image_uuid': fields.StringField(nullable=True),

        'key_name': fields.StringField(nullable=True),

        'kernel_id': fields.StringField(nullable=True),
        'ramdisk_id': fields.StringField(nullable=True),

        'created_at': fields.DateTimeField(nullable=True),
        'launched_at': fields.DateTimeField(nullable=True),
        'terminated_at': fields.DateTimeField(nullable=True),
        'deleted_at': fields.DateTimeField(nullable=True),
        'updated_at': fields.DateTimeField(nullable=True),

        'state': fields.InstanceStateField(nullable=True),
        'power_state': fields.InstancePowerStateField(nullable=True),
        'task_state': fields.InstanceTaskStateField(nullable=True),
        'progress': fields.IntegerField(nullable=True),

        'ip_addresses': fields.ListOfObjectsField('IpPayload'),
        'block_devices': fields.ListOfObjectsField('BlockDevicePayload',
                                                   nullable=True),

        'metadata': fields.DictOfStringsField(),
        'locked': fields.BooleanField(),
        'auto_disk_config': fields.DiskConfigField(),

        'request_id': fields.StringField(nullable=True),
        'action_initiator_user': fields.StringField(nullable=True),
        'action_initiator_project': fields.StringField(nullable=True),
        'locked_reason': fields.StringField(nullable=True),
    }

    def __init__(self, context, instance, bdms=None):
        super(InstancePayload, self).__init__()
        network_info = instance.get_network_info()
        self.ip_addresses = IpPayload.from_network_info(network_info)
        self.flavor = flavor_payload.FlavorPayload(flavor=instance.flavor)
        if bdms is not None:
            self.block_devices = BlockDevicePayload.from_bdms(bdms)
        else:
            self.block_devices = BlockDevicePayload.from_instance(instance)
        # NOTE(Kevin_Zheng): Don't include request_id for periodic tasks,
        # RequestContext for periodic tasks does not include project_id
        # and user_id. Consider modify this once periodic tasks got a
        # consistent request_id.
        self.request_id = context.request_id if (context.project_id and
                                                 context.user_id) else None
        self.action_initiator_user = context.user_id
        self.action_initiator_project = context.project_id
        self.locked_reason = instance.system_metadata.get("locked_reason")
        self.populate_schema(instance=instance)


@nova_base.NovaObjectRegistry.register_notification
class InstanceActionPayload(InstancePayload):
    # No SCHEMA as all the additional fields are calculated

    # Version 1.1: locked and display_description added to InstancePayload
    # Version 1.2: Added auto_disk_config field to InstancePayload
    # Version 1.3: Added key_name field to InstancePayload
    # Version 1.4: Add BDM related data
    # Version 1.5: Added updated_at field to InstancePayload
    # Version 1.6: Added request_id field to InstancePayload
    # Version 1.7: Added action_initiator_user and action_initiator_project to
    #              InstancePayload
    # Version 1.8: Added locked_reason field to InstancePayload
    VERSION = '1.8'
    fields = {
        'fault': fields.ObjectField('ExceptionPayload', nullable=True),
        'request_id': fields.StringField(nullable=True),
    }

    def __init__(self, context, instance, fault, bdms=None):
        super(InstanceActionPayload, self).__init__(context=context,
                                                    instance=instance,
                                                    bdms=bdms)
        self.fault = fault


@nova_base.NovaObjectRegistry.register_notification
class InstanceActionVolumePayload(InstanceActionPayload):
    # Version 1.0: Initial version
    # Version 1.1: Added key_name field to InstancePayload
    # Version 1.2: Add BDM related data
    # Version 1.3: Added updated_at field to InstancePayload
    # Version 1.4: Added request_id field to InstancePayload
    # Version 1.5: Added action_initiator_user and action_initiator_project to
    #              InstancePayload
    # Version 1.6: Added locked_reason field to InstancePayload
    VERSION = '1.6'
    fields = {
        'volume_id': fields.UUIDField()
    }

    def __init__(self, context, instance, fault, volume_id):
        super(InstanceActionVolumePayload, self).__init__(
                context=context,
                instance=instance,
                fault=fault)
        self.volume_id = volume_id


@nova_base.NovaObjectRegistry.register_notification
class InstanceActionVolumeSwapPayload(InstanceActionPayload):
    # No SCHEMA as all the additional fields are calculated

    # Version 1.1: locked and display_description added to InstancePayload
    # Version 1.2: Added auto_disk_config field to InstancePayload
    # Version 1.3: Added key_name field to InstancePayload
    # Version 1.4: Add BDM related data
    # Version 1.5: Added updated_at field to InstancePayload
    # Version 1.6: Added request_id field to InstancePayload
    # Version 1.7: Added action_initiator_user and action_initiator_project to
    #              InstancePayload
    # Version 1.8: Added locked_reason field to InstancePayload
    VERSION = '1.8'
    fields = {
        'old_volume_id': fields.UUIDField(),
        'new_volume_id': fields.UUIDField(),
    }

    def __init__(self, context, instance, fault, old_volume_id, new_volume_id):
        super(InstanceActionVolumeSwapPayload, self).__init__(
                context=context,
                instance=instance,
                fault=fault)
        self.old_volume_id = old_volume_id
        self.new_volume_id = new_volume_id


@nova_base.NovaObjectRegistry.register_notification
class InstanceCreatePayload(InstanceActionPayload):
    # No SCHEMA as all the additional fields are calculated

    # Version 1.2: Initial version. It starts at 1.2 to match with the version
    #              of the InstanceActionPayload at the time when this specific
    #              payload is created as a child of it so that the
    #              instance.create notification using this new payload does not
    #              have decreasing version.
    #         1.3: Add keypairs field
    #         1.4: Add key_name field to InstancePayload
    #         1.5: Add BDM related data to InstancePayload
    #         1.6: Add tags field to InstanceCreatePayload
    #         1.7: Added updated_at field to InstancePayload
    #         1.8: Added request_id field to InstancePayload
    #         1.9: Add trusted_image_certificates field to
    #              InstanceCreatePayload
    #         1.10: Added action_initiator_user and action_initiator_project to
    #              InstancePayload
    #         1.11: Added instance_name to InstanceCreatePayload
    # Version 1.12: Added locked_reason field to InstancePayload
    VERSION = '1.12'
    fields = {
        'keypairs': fields.ListOfObjectsField('KeypairPayload'),
        'tags': fields.ListOfStringsField(),
        'trusted_image_certificates': fields.ListOfStringsField(
            nullable=True),
        'instance_name': fields.StringField(nullable=True),
    }

    def __init__(self, context, instance, fault, bdms):
        super(InstanceCreatePayload, self).__init__(
            context=context,
            instance=instance,
            fault=fault,
            bdms=bdms)
        self.keypairs = [keypair_payload.KeypairPayload(keypair=keypair)
                         for keypair in instance.keypairs]
        self.tags = [instance_tag.tag
                     for instance_tag in instance.tags]
        self.trusted_image_certificates = None
        if instance.trusted_certs:
            self.trusted_image_certificates = instance.trusted_certs.ids
        self.instance_name = instance.name


@nova_base.NovaObjectRegistry.register_notification
class InstanceActionResizePrepPayload(InstanceActionPayload):
    # No SCHEMA as all the additional fields are calculated

    # Version 1.0: Initial version
    # Version 1.1: Added request_id field to InstancePayload
    # Version 1.2: Added action_initiator_user and action_initiator_project to
    #              InstancePayload
    # Version 1.3: Added locked_reason field to InstancePayload
    VERSION = '1.3'
    fields = {
        'new_flavor': fields.ObjectField('FlavorPayload', nullable=True)
    }

    def __init__(self, context, instance, fault, new_flavor):
        super(InstanceActionResizePrepPayload, self).__init__(
                context=context,
                instance=instance,
                fault=fault)
        self.new_flavor = new_flavor


@nova_base.NovaObjectRegistry.register_notification
class InstanceUpdatePayload(InstancePayload):
    # Version 1.0: Initial version
    # Version 1.1: locked and display_description added to InstancePayload
    # Version 1.2: Added tags field
    # Version 1.3: Added auto_disk_config field to InstancePayload
    # Version 1.4: Added key_name field to InstancePayload
    # Version 1.5: Add BDM related data
    # Version 1.6: Added updated_at field to InstancePayload
    # Version 1.7: Added request_id field to InstancePayload
    # Version 1.8: Added action_initiator_user and action_initiator_project to
    #              InstancePayload
    # Version 1.9: Added locked_reason field to InstancePayload
    VERSION = '1.9'
    fields = {
        'state_update': fields.ObjectField('InstanceStateUpdatePayload'),
        'audit_period': fields.ObjectField('AuditPeriodPayload'),
        'bandwidth': fields.ListOfObjectsField('BandwidthPayload'),
        'old_display_name': fields.StringField(nullable=True),
        'tags': fields.ListOfStringsField(),
    }

    def __init__(self, context, instance, state_update, audit_period,
                 bandwidth, old_display_name):
        super(InstanceUpdatePayload, self).__init__(
            context=context, instance=instance)
        self.state_update = state_update
        self.audit_period = audit_period
        self.bandwidth = bandwidth
        self.old_display_name = old_display_name
        self.tags = [instance_tag.tag
                     for instance_tag in instance.tags.objects]


@nova_base.NovaObjectRegistry.register_notification
class InstanceActionRescuePayload(InstanceActionPayload):
    # Version 1.0: Initial version
    # Version 1.1: Added request_id field to InstancePayload
    # Version 1.2: Added action_initiator_user and action_initiator_project to
    #              InstancePayload
    # Version 1.3: Added locked_reason field to InstancePayload
    VERSION = '1.3'
    fields = {
        'rescue_image_ref': fields.UUIDField(nullable=True)
    }

    def __init__(self, context, instance, fault, rescue_image_ref):
        super(InstanceActionRescuePayload, self).__init__(
                context=context,
                instance=instance,
                fault=fault)
        self.rescue_image_ref = rescue_image_ref


@nova_base.NovaObjectRegistry.register_notification
class InstanceActionRebuildPayload(InstanceActionPayload):
    # No SCHEMA as all the additional fields are calculated

    # Version 1.7: Initial version. It starts at 1.7 to equal one more than
    #              the version of the InstanceActionPayload at the time
    #              when this specific payload is created so that the
    #              instance.rebuild.* notifications using this new payload
    #              signal the change of nova_object.name.
    # Version 1.8: Added action_initiator_user and action_initiator_project to
    #              InstancePayload
    # Version 1.9: Added locked_reason field to InstancePayload
    VERSION = '1.9'
    fields = {
        'trusted_image_certificates': fields.ListOfStringsField(
            nullable=True)
    }

    def __init__(self, context, instance, fault, bdms=None):
        super(InstanceActionRebuildPayload, self).__init__(
                context=context,
                instance=instance,
                fault=fault,
                bdms=bdms)
        self.trusted_image_certificates = None
        if instance.trusted_certs:
            self.trusted_image_certificates = instance.trusted_certs.ids


@nova_base.NovaObjectRegistry.register_notification
class IpPayload(base.NotificationPayloadBase):
    # Version 1.0: Initial version
    VERSION = '1.0'
    fields = {
        'label': fields.StringField(),
        'mac': fields.MACAddressField(),
        'meta': fields.DictOfStringsField(),
        'port_uuid': fields.UUIDField(nullable=True),
        'version': fields.IntegerField(),
        'address': fields.IPV4AndV6AddressField(),
        'device_name': fields.StringField(nullable=True)
    }

    def __init__(self, label, mac, meta, port_uuid, version, address,
                 device_name):
        super(IpPayload, self).__init__()
        self.label = label
        self.mac = mac
        self.meta = meta
        self.port_uuid = port_uuid
        self.version = version
        self.address = address
        self.device_name = device_name

    @classmethod
    def from_network_info(cls, network_info):
        """Returns a list of IpPayload object based on the passed
        network_info.
        """
        ips = []
        if network_info is not None:
            for vif in network_info:
                for ip in vif.fixed_ips():
                    ips.append(cls(
                        label=vif["network"]["label"],
                        mac=vif["address"],
                        meta=vif["meta"],
                        port_uuid=vif["id"],
                        version=ip["version"],
                        address=ip["address"],
                        device_name=vif["devname"]))
        return ips


@nova_base.NovaObjectRegistry.register_notification
class BandwidthPayload(base.NotificationPayloadBase):
    # Version 1.0: Initial version
    VERSION = '1.0'
    fields = {
        'network_name': fields.StringField(),
        'in_bytes': fields.IntegerField(),
        'out_bytes': fields.IntegerField(),
    }

    def __init__(self, network_name, in_bytes, out_bytes):
        super(BandwidthPayload, self).__init__()
        self.network_name = network_name
        self.in_bytes = in_bytes
        self.out_bytes = out_bytes


@nova_base.NovaObjectRegistry.register_notification
class AuditPeriodPayload(base.NotificationPayloadBase):
    # Version 1.0: Initial version
    VERSION = '1.0'
    fields = {
        'audit_period_beginning': fields.DateTimeField(),
        'audit_period_ending': fields.DateTimeField(),
    }

    def __init__(self, audit_period_beginning, audit_period_ending):
        super(AuditPeriodPayload, self).__init__()
        self.audit_period_beginning = audit_period_beginning
        self.audit_period_ending = audit_period_ending


@nova_base.NovaObjectRegistry.register_notification
class BlockDevicePayload(base.NotificationPayloadBase):
    # Version 1.0: Initial version
    VERSION = '1.0'

    SCHEMA = {
        'device_name': ('bdm', 'device_name'),
        'boot_index': ('bdm', 'boot_index'),
        'delete_on_termination': ('bdm', 'delete_on_termination'),
        'volume_id': ('bdm', 'volume_id'),
        'tag': ('bdm', 'tag')
    }

    fields = {
        'device_name': fields.StringField(nullable=True),
        'boot_index': fields.IntegerField(nullable=True),
        'delete_on_termination': fields.BooleanField(default=False),
        'volume_id': fields.UUIDField(),
        'tag': fields.StringField(nullable=True)
    }

    def __init__(self, bdm):
        super(BlockDevicePayload, self).__init__()
        self.populate_schema(bdm=bdm)

    @classmethod
    def from_instance(cls, instance):
        """Returns a list of BlockDevicePayload objects based on the passed
        bdms.
        """
        if not CONF.notifications.bdms_in_notifications:
            return None

        instance_bdms = instance.get_bdms()
        if instance_bdms is not None:
            return cls.from_bdms(instance_bdms)
        else:
            return []

    @classmethod
    def from_bdms(cls, bdms):
        """Returns a list of BlockDevicePayload objects based on the passed
        BlockDeviceMappingList.
        """
        payloads = []
        for bdm in bdms:
            if bdm.volume_id is not None:
                payloads.append(cls(bdm))
        return payloads


@nova_base.NovaObjectRegistry.register_notification
class InstanceStateUpdatePayload(base.NotificationPayloadBase):
    # Version 1.0: Initial version
    VERSION = '1.0'
    fields = {
        'old_state': fields.StringField(nullable=True),
        'state': fields.StringField(nullable=True),
        'old_task_state': fields.StringField(nullable=True),
        'new_task_state': fields.StringField(nullable=True),
    }

    def __init__(self, old_state, state, old_task_state, new_task_state):
        super(InstanceStateUpdatePayload, self).__init__()
        self.old_state = old_state
        self.state = state
        self.old_task_state = old_task_state
        self.new_task_state = new_task_state


@base.notification_sample('instance-delete-start.json')
@base.notification_sample('instance-delete-end.json')
@base.notification_sample('instance-pause-start.json')
@base.notification_sample('instance-pause-end.json')
@base.notification_sample('instance-unpause-start.json')
@base.notification_sample('instance-unpause-end.json')
@base.notification_sample('instance-resize-start.json')
@base.notification_sample('instance-resize-end.json')
@base.notification_sample('instance-resize-error.json')
@base.notification_sample('instance-suspend-start.json')
@base.notification_sample('instance-suspend-end.json')
@base.notification_sample('instance-power_on-start.json')
@base.notification_sample('instance-power_on-end.json')
@base.notification_sample('instance-power_off-start.json')
@base.notification_sample('instance-power_off-end.json')
@base.notification_sample('instance-reboot-start.json')
@base.notification_sample('instance-reboot-end.json')
@base.notification_sample('instance-reboot-error.json')
@base.notification_sample('instance-shutdown-start.json')
@base.notification_sample('instance-shutdown-end.json')
@base.notification_sample('instance-interface_attach-start.json')
@base.notification_sample('instance-interface_attach-end.json')
@base.notification_sample('instance-interface_attach-error.json')
@base.notification_sample('instance-shelve-start.json')
@base.notification_sample('instance-shelve-end.json')
@base.notification_sample('instance-resume-start.json')
@base.notification_sample('instance-resume-end.json')
@base.notification_sample('instance-restore-start.json')
@base.notification_sample('instance-restore-end.json')
@base.notification_sample('instance-evacuate.json')
@base.notification_sample('instance-resize_finish-start.json')
@base.notification_sample('instance-resize_finish-end.json')
@base.notification_sample('instance-live_migration_pre-start.json')
@base.notification_sample('instance-live_migration_pre-end.json')
@base.notification_sample('instance-live_migration_abort-start.json')
@base.notification_sample('instance-live_migration_abort-end.json')
@base.notification_sample('instance-live_migration_post-start.json')
@base.notification_sample('instance-live_migration_post-end.json')
@base.notification_sample('instance-live_migration_post_dest-start.json')
@base.notification_sample('instance-live_migration_post_dest-end.json')
@base.notification_sample('instance-live_migration_rollback-start.json')
@base.notification_sample('instance-live_migration_rollback-end.json')
@base.notification_sample('instance-live_migration_rollback_dest-start.json')
@base.notification_sample('instance-live_migration_rollback_dest-end.json')
@base.notification_sample('instance-interface_detach-start.json')
@base.notification_sample('instance-interface_detach-end.json')
@base.notification_sample('instance-resize_confirm-start.json')
@base.notification_sample('instance-resize_confirm-end.json')
@base.notification_sample('instance-resize_revert-start.json')
@base.notification_sample('instance-resize_revert-end.json')
@base.notification_sample('instance-live_migration_force_complete-start.json')
@base.notification_sample('instance-live_migration_force_complete-end.json')
@base.notification_sample('instance-shelve_offload-start.json')
@base.notification_sample('instance-shelve_offload-end.json')
@base.notification_sample('instance-soft_delete-start.json')
@base.notification_sample('instance-soft_delete-end.json')
@base.notification_sample('instance-trigger_crash_dump-start.json')
@base.notification_sample('instance-trigger_crash_dump-end.json')
@base.notification_sample('instance-unrescue-start.json')
@base.notification_sample('instance-unrescue-end.json')
@base.notification_sample('instance-unshelve-start.json')
@base.notification_sample('instance-unshelve-end.json')
@base.notification_sample('instance-lock.json')
@base.notification_sample('instance-unlock.json')
@nova_base.NovaObjectRegistry.register_notification
class InstanceActionNotification(base.NotificationBase):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'payload': fields.ObjectField('InstanceActionPayload')
    }


@base.notification_sample('instance-update.json')
@nova_base.NovaObjectRegistry.register_notification
class InstanceUpdateNotification(base.NotificationBase):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'payload': fields.ObjectField('InstanceUpdatePayload')
    }


@base.notification_sample('instance-volume_swap-start.json')
@base.notification_sample('instance-volume_swap-end.json')
@base.notification_sample('instance-volume_swap-error.json')
@nova_base.NovaObjectRegistry.register_notification
class InstanceActionVolumeSwapNotification(base.NotificationBase):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'payload': fields.ObjectField('InstanceActionVolumeSwapPayload')
    }


@base.notification_sample('instance-volume_attach-start.json')
@base.notification_sample('instance-volume_attach-end.json')
@base.notification_sample('instance-volume_attach-error.json')
@base.notification_sample('instance-volume_detach-start.json')
@base.notification_sample('instance-volume_detach-end.json')
@nova_base.NovaObjectRegistry.register_notification
class InstanceActionVolumeNotification(base.NotificationBase):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'payload': fields.ObjectField('InstanceActionVolumePayload')
    }


@base.notification_sample('instance-create-start.json')
@base.notification_sample('instance-create-end.json')
@base.notification_sample('instance-create-error.json')
@nova_base.NovaObjectRegistry.register_notification
class InstanceCreateNotification(base.NotificationBase):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'payload': fields.ObjectField('InstanceCreatePayload')
    }


@base.notification_sample('instance-resize_prep-start.json')
@base.notification_sample('instance-resize_prep-end.json')
@nova_base.NovaObjectRegistry.register_notification
class InstanceActionResizePrepNotification(base.NotificationBase):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'payload': fields.ObjectField('InstanceActionResizePrepPayload')
    }


@base.notification_sample('instance-snapshot-start.json')
@base.notification_sample('instance-snapshot-end.json')
@nova_base.NovaObjectRegistry.register_notification
class InstanceActionSnapshotNotification(base.NotificationBase):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'payload': fields.ObjectField('InstanceActionSnapshotPayload')
    }


@base.notification_sample('instance-rescue-start.json')
@base.notification_sample('instance-rescue-end.json')
@nova_base.NovaObjectRegistry.register_notification
class InstanceActionRescueNotification(base.NotificationBase):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'payload': fields.ObjectField('InstanceActionRescuePayload')
    }


@base.notification_sample('instance-rebuild_scheduled.json')
@base.notification_sample('instance-rebuild-start.json')
@base.notification_sample('instance-rebuild-end.json')
@base.notification_sample('instance-rebuild-error.json')
@nova_base.NovaObjectRegistry.register_notification
class InstanceActionRebuildNotification(base.NotificationBase):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'payload': fields.ObjectField('InstanceActionRebuildPayload')
    }


@nova_base.NovaObjectRegistry.register_notification
class InstanceActionSnapshotPayload(InstanceActionPayload):
    # Version 1.6: Initial version. It starts at version 1.6 as
    #              instance.snapshot.start and .end notifications are switched
    #              from using InstanceActionPayload 1.5 to this new payload and
    #              also it added a new field so we wanted to keep the version
    #              number increasing to signal the change.
    # Version 1.7: Added request_id field to InstancePayload
    # Version 1.8: Added action_initiator_user and action_initiator_project to
    #              InstancePayload
    # Version 1.9: Added locked_reason field to InstancePayload
    VERSION = '1.9'
    fields = {
        'snapshot_image_id': fields.UUIDField(),
    }

    def __init__(self, context, instance, fault, snapshot_image_id):
        super(InstanceActionSnapshotPayload, self).__init__(
                context=context,
                instance=instance,
                fault=fault)
        self.snapshot_image_id = snapshot_image_id


@nova_base.NovaObjectRegistry.register_notification
class InstanceExistsPayload(InstancePayload):
    # Version 1.0: Initial version
    # Version 1.1: Added action_initiator_user and action_initiator_project to
    #              InstancePayload
    # Version 1.2: Added locked_reason field to InstancePayload
    VERSION = '1.2'
    fields = {
        'audit_period': fields.ObjectField('AuditPeriodPayload'),
        'bandwidth': fields.ListOfObjectsField('BandwidthPayload'),
    }

    def __init__(self, context, instance, audit_period, bandwidth):
        super(InstanceExistsPayload, self).__init__(context=context,
                                                    instance=instance)
        self.audit_period = audit_period
        self.bandwidth = bandwidth


@base.notification_sample('instance-exists.json')
@nova_base.NovaObjectRegistry.register_notification
class InstanceExistsNotification(base.NotificationBase):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'payload': fields.ObjectField('InstanceExistsPayload')
    }



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from migrate.changeset.constraint import ForeignKeyConstraint
from migrate import UniqueConstraint
from sqlalchemy import Column
from sqlalchemy import DateTime
from sqlalchemy import Index
from sqlalchemy import Integer
from sqlalchemy import MetaData
from sqlalchemy import String
from sqlalchemy import Table


def upgrade(migrate_engine):
    meta = MetaData()
    meta.bind = migrate_engine

    cell_mappings = Table('cell_mappings', meta, autoload=True)
    instance_mappings = Table('instance_mappings', meta,
        Column('created_at', DateTime),
        Column('updated_at', DateTime),
        Column('id', Integer, primary_key=True, nullable=False),
        Column('instance_uuid', String(length=36), nullable=False),
        Column('cell_id', Integer, nullable=False),
        Column('project_id', String(length=255), nullable=False),
        UniqueConstraint('instance_uuid',
            name='uniq_instance_mappings0instance_uuid'),
        Index('instance_uuid_idx', 'instance_uuid'),
        Index('project_id_idx', 'project_id'),
        ForeignKeyConstraint(columns=['cell_id'],
            refcolumns=[cell_mappings.c.id]),
        mysql_engine='InnoDB',
        mysql_charset='utf8'
    )

    instance_mappings.create(checkfirst=True)



# (c) Copyright 2015 - 2019  StorPool
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from os_brick import initiator
from os_brick.initiator import connector
from oslo_log import log as logging

from nova import utils
from nova.virt.libvirt.volume import volume as libvirt_volume

LOG = logging.getLogger(__name__)


class LibvirtStorPoolVolumeDriver(libvirt_volume.LibvirtVolumeDriver):
    """Driver to attach StorPool volumes to libvirt."""

    def __init__(self, host):
        super(LibvirtStorPoolVolumeDriver, self).__init__(host)

        self.connector = connector.InitiatorConnector.factory(
            initiator.STORPOOL, utils.get_root_helper())

    def get_config(self, connection_info, disk_info):
        """Returns xml for libvirt."""
        conf = super(LibvirtStorPoolVolumeDriver,
                     self).get_config(connection_info, disk_info)
        conf.driver_io = "native"
        return conf

    def connect_volume(self, connection_info, instance):
        LOG.debug("Attaching StorPool volume %s",
                  connection_info['data']['volume'], instance=instance)
        device_info = self.connector.connect_volume(connection_info['data'])
        LOG.debug("Attached StorPool volume %s",
                  device_info, instance=instance)
        connection_info['data']['device_path'] = device_info['path']

    def disconnect_volume(self, connection_info, instance):
        LOG.debug("Detaching StorPool volume %s",
                  connection_info['data']['volume'], instance=instance)
        self.connector.disconnect_volume(connection_info['data'], None)
        LOG.debug("Detached StorPool volume", instance=instance)

    def extend_volume(self, connection_info, instance, requested_size):
        """Extend the volume."""
        LOG.debug("Extending StorPool volume %s",
                  connection_info['data']['volume'], instance=instance)
        new_size = self.connector.extend_volume(connection_info['data'])
        LOG.debug("Extended StorPool Volume %s; new_size=%s",
                  connection_info['data']['device_path'],
                  new_size, instance=instance)
        return new_size



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import fixtures
from oslo_utils.fixture import uuidsentinel as uuids

from nova.api.openstack.compute import server_topology
from nova.compute import vm_states
from nova import objects
from nova.policies import server_topology as policies
from nova.tests.unit.api.openstack import fakes
from nova.tests.unit import fake_instance
from nova.tests.unit.policies import base


class ServerTopologyPolicyTest(base.BasePolicyTest):
    """Test Server Topology APIs policies with all possible context.
    This class defines the set of context with different roles
    which are allowed and not allowed to pass the policy checks.
    With those set of context, it will call the API operation and
    verify the expected behaviour.
    """

    def setUp(self):
        super(ServerTopologyPolicyTest, self).setUp()
        self.controller = server_topology.ServerTopologyController()
        self.req = fakes.HTTPRequest.blank('', version='2.78')
        self.mock_get = self.useFixture(
            fixtures.MockPatch('nova.api.openstack.common.get_instance')).mock
        self.instance = fake_instance.fake_instance_obj(
                self.project_member_context,
                id=1, uuid=uuids.fake_id, project_id=self.project_id,
                vm_state=vm_states.ACTIVE)
        self.mock_get.return_value = self.instance
        self.instance.numa_topology = objects.InstanceNUMATopology(
                instance_uuid = self.instance.uuid,
                cells=[objects.InstanceNUMACell(
                    node=0, memory=1024, pagesize=4, id=123,
                    cpu_topology=None,
                    cpu_pinning={},
                    cpuset=set([0, 1]))])

        # Check that system reader or and server owner is able to get
        # the server topology.
        self.system_reader_or_owner_authorized_contexts = [
            self.legacy_admin_context, self.system_admin_context,
            self.project_admin_context, self.project_member_context,
            self.project_reader_context, self.project_foo_context,
            self.system_member_context, self.system_reader_context]
        # Check that non-stem reader/owner is not able to get
        # the server topology.
        self.system_reader_or_owner_unauthorized_contexts = [
            self.system_foo_context, self.other_project_member_context,
            self.other_project_reader_context,
        ]
        # Check that system reader is able to get the server topology
        # host information.
        self.system_reader_authorized_contexts = [
            self.legacy_admin_context, self.system_admin_context,
            self.project_admin_context, self.system_member_context,
            self.system_reader_context]
        # Check that non-system reader is not able to get the server topology
        # host information.
        self.system_reader_unauthorized_contexts = [
            self.system_foo_context, self.project_member_context,
            self.other_project_member_context,
            self.project_foo_context, self.project_reader_context,
            self.other_project_reader_context
        ]

    def test_index_server_topology_policy(self):
        rule_name = policies.BASE_POLICY_NAME % 'index'
        self.common_policy_check(
            self.system_reader_or_owner_authorized_contexts,
            self.system_reader_or_owner_unauthorized_contexts,
            rule_name,
            self.controller.index,
            self.req, self.instance.uuid)

    def test_index_host_server_topology_policy(self):
        rule_name = policies.BASE_POLICY_NAME % 'host:index'
        # 'index' policy is checked before 'host:index' so
        # we have to allow it for everyone otherwise it will
        # fail first for unauthorized contexts.
        rule = policies.BASE_POLICY_NAME % 'index'
        self.policy.set_rules({rule: "@"}, overwrite=False)
        authorize_res, unauthorize_res = self.common_policy_check(
            self.system_reader_authorized_contexts,
            self.system_reader_unauthorized_contexts,
            rule_name, self.controller.index, self.req, self.instance.uuid,
            fatal=False)
        for resp in authorize_res:
            self.assertEqual(123, resp['nodes'][0]['host_node'])
            self.assertEqual({}, resp['nodes'][0]['cpu_pinning'])
        for resp in unauthorize_res:
            self.assertNotIn('host_node', resp['nodes'][0])
            self.assertNotIn('cpu_pinning', resp['nodes'][0])


class ServerTopologyScopeTypePolicyTest(ServerTopologyPolicyTest):
    """Test Server Topology APIs policies with system scope enabled.
    This class set the nova.conf [oslo_policy] enforce_scope to True
    so that we can switch on the scope checking on oslo policy side.
    It defines the set of context with scoped token
    which are allowed and not allowed to pass the policy checks.
    With those set of context, it will run the API operation and
    verify the expected behaviour.
    """

    def setUp(self):
        super(ServerTopologyScopeTypePolicyTest, self).setUp()
        self.flags(enforce_scope=True, group="oslo_policy")

        # Check that system reader is able to get the server topology
        # host information.
        self.system_reader_authorized_contexts = [
            self.system_admin_context, self.system_member_context,
            self.system_reader_context]
        # Check that non-system/reader is not able to get the server topology
        # host information.
        self.system_reader_unauthorized_contexts = [
            self.legacy_admin_context, self.system_foo_context,
            self.project_admin_context, self.project_member_context,
            self.other_project_member_context,
            self.project_foo_context, self.project_reader_context,
            self.other_project_reader_context,
        ]


class ServerTopologyNoLegacyPolicyTest(ServerTopologyScopeTypePolicyTest):
    """Test Server Topology APIs policies with system scope enabled,
    and no more deprecated rules that allow the legacy admin API to
    access system APIs.
    """
    without_deprecated_rules = True

    def setUp(self):
        super(ServerTopologyNoLegacyPolicyTest, self).setUp()
        # Check that system reader/owner is able to get
        # the server topology.
        self.system_reader_or_owner_authorized_contexts = [
            self.system_admin_context,
            self.project_admin_context, self.project_member_context,
            self.system_member_context, self.system_reader_context,
            self.project_reader_context]
        # Check that non-system/reader/owner is not able to get
        # the server topology.
        self.system_reader_or_owner_unauthorized_contexts = [
            self.legacy_admin_context, self.system_foo_context,
            self.other_project_member_context, self.project_foo_context,
            self.other_project_reader_context,
        ]



#    Copyright 2018 NTT Corporation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from nova.notifications.objects import base
from nova.objects import base as nova_base
from nova.objects import fields


@base.notification_sample('metrics-update.json')
@nova_base.NovaObjectRegistry.register_notification
class MetricsNotification(base.NotificationBase):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'payload': fields.ObjectField('MetricsPayload')
    }


@nova_base.NovaObjectRegistry.register_notification
class MetricPayload(base.NotificationPayloadBase):
    # Version 1.0: Initial version
    VERSION = '1.0'

    SCHEMA = {
        'name': ('monitor_metric', 'name'),
        'value': ('monitor_metric', 'value'),
        'numa_membw_values': ('monitor_metric', 'numa_membw_values'),
        'timestamp': ('monitor_metric', 'timestamp'),
        'source': ('monitor_metric', 'source'),
    }

    fields = {
        'name': fields.MonitorMetricTypeField(),
        'value': fields.IntegerField(),
        'numa_membw_values': fields.DictOfIntegersField(nullable=True),
        'timestamp': fields.DateTimeField(),
        'source': fields.StringField(),
    }

    def __init__(self, monitor_metric):
        super(MetricPayload, self).__init__()
        self.populate_schema(monitor_metric=monitor_metric)

    @classmethod
    def from_monitor_metric_list_obj(cls, monitor_metric_list):
        """Returns a list of MetricPayload objects based on the passed
        MonitorMetricList object.
        """
        payloads = []
        for monitor_metric in monitor_metric_list:
            payloads.append(cls(monitor_metric))
        return payloads


@nova_base.NovaObjectRegistry.register_notification
class MetricsPayload(base.NotificationPayloadBase):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'host': fields.StringField(),
        'host_ip': fields.StringField(),
        'nodename': fields.StringField(),
        'metrics': fields.ListOfObjectsField('MetricPayload'),
    }

    def __init__(self, host, host_ip, nodename, monitor_metric_list):
        super(MetricsPayload, self).__init__()
        self.host = host
        self.host_ip = host_ip
        self.nodename = nodename
        self.metrics = MetricPayload.from_monitor_metric_list_obj(
            monitor_metric_list)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import re
import string

import six

from nova import exception
from nova.i18n import _
from nova.pci import utils

MAX_VENDOR_ID = 0xFFFF
MAX_PRODUCT_ID = 0xFFFF
MAX_FUNC = 0x7
MAX_DOMAIN = 0xFFFF
MAX_BUS = 0xFF
MAX_SLOT = 0x1F
ANY = '*'
REGEX_ANY = '.*'


@six.add_metaclass(abc.ABCMeta)
class PciAddressSpec(object):
    """Abstract class for all PCI address spec styles

    This class checks the address fields of the pci.passthrough_whitelist
    """

    @abc.abstractmethod
    def match(self, pci_addr):
        pass

    def is_single_address(self):
        return all([
            all(c in string.hexdigits for c in self.domain),
            all(c in string.hexdigits for c in self.bus),
            all(c in string.hexdigits for c in self.slot),
            all(c in string.hexdigits for c in self.func)])

    def _set_pci_dev_info(self, prop, maxval, hex_value):
        a = getattr(self, prop)
        if a == ANY:
            return
        try:
            v = int(a, 16)
        except ValueError:
            raise exception.PciConfigInvalidWhitelist(
                reason=_("property %(property)s ('%(attr)s') does not parse "
                         "as a hex number.") % {'property': prop, 'attr': a})
        if v > maxval:
            raise exception.PciConfigInvalidWhitelist(
                reason=_("property %(property)s (%(attr)s) is greater than "
                         "the maximum allowable value (%(max)X).") %
                         {'property': prop, 'attr': a, 'max': maxval})
        setattr(self, prop, hex_value % v)


class PhysicalPciAddress(PciAddressSpec):
    """Manages the address fields for a fully-qualified PCI address.

    This function class will validate the address fields for a single
    PCI device.
    """
    def __init__(self, pci_addr):
        try:
            if isinstance(pci_addr, dict):
                self.domain = pci_addr['domain']
                self.bus = pci_addr['bus']
                self.slot = pci_addr['slot']
                self.func = pci_addr['function']
            else:
                self.domain, self.bus, self.slot, self.func = (
                    utils.get_pci_address_fields(pci_addr))
            self._set_pci_dev_info('func', MAX_FUNC, '%1x')
            self._set_pci_dev_info('domain', MAX_DOMAIN, '%04x')
            self._set_pci_dev_info('bus', MAX_BUS, '%02x')
            self._set_pci_dev_info('slot', MAX_SLOT, '%02x')
        except (KeyError, ValueError):
            raise exception.PciDeviceWrongAddressFormat(address=pci_addr)

    def match(self, phys_pci_addr):
        conditions = [
            self.domain == phys_pci_addr.domain,
            self.bus == phys_pci_addr.bus,
            self.slot == phys_pci_addr.slot,
            self.func == phys_pci_addr.func,
            ]
        return all(conditions)


class PciAddressGlobSpec(PciAddressSpec):
    """Manages the address fields with glob style.

    This function class will validate the address fields with glob style,
    check for wildcards, and insert wildcards where the field is left blank.
    """

    def __init__(self, pci_addr):
        self.domain = ANY
        self.bus = ANY
        self.slot = ANY
        self.func = ANY

        dbs, sep, func = pci_addr.partition('.')
        if func:
            self.func = func.strip()
            self._set_pci_dev_info('func', MAX_FUNC, '%01x')
        if dbs:
            dbs_fields = dbs.split(':')
            if len(dbs_fields) > 3:
                raise exception.PciDeviceWrongAddressFormat(address=pci_addr)
            # If we got a partial address like ":00.", we need to turn this
            # into a domain of ANY, a bus of ANY, and a slot of 00. This code
            # allows the address bus and/or domain to be left off
            dbs_all = [ANY] * (3 - len(dbs_fields))
            dbs_all.extend(dbs_fields)
            dbs_checked = [s.strip() or ANY for s in dbs_all]
            self.domain, self.bus, self.slot = dbs_checked
            self._set_pci_dev_info('domain', MAX_DOMAIN, '%04x')
            self._set_pci_dev_info('bus', MAX_BUS, '%02x')
            self._set_pci_dev_info('slot', MAX_SLOT, '%02x')

    def match(self, phys_pci_addr):
        conditions = [
            self.domain in (ANY, phys_pci_addr.domain),
            self.bus in (ANY, phys_pci_addr.bus),
            self.slot in (ANY, phys_pci_addr.slot),
            self.func in (ANY, phys_pci_addr.func)
            ]
        return all(conditions)


class PciAddressRegexSpec(PciAddressSpec):
    """Manages the address fields with regex style.

    This function class will validate the address fields with regex style.
    The validation includes check for all PCI address attributes and validate
    their regex.
    """
    def __init__(self, pci_addr):
        try:
            self.domain = pci_addr.get('domain', REGEX_ANY)
            self.bus = pci_addr.get('bus', REGEX_ANY)
            self.slot = pci_addr.get('slot', REGEX_ANY)
            self.func = pci_addr.get('function', REGEX_ANY)
            self.domain_regex = re.compile(self.domain)
            self.bus_regex = re.compile(self.bus)
            self.slot_regex = re.compile(self.slot)
            self.func_regex = re.compile(self.func)
        except re.error:
            raise exception.PciDeviceWrongAddressFormat(address=pci_addr)

    def match(self, phys_pci_addr):
        conditions = [
            bool(self.domain_regex.match(phys_pci_addr.domain)),
            bool(self.bus_regex.match(phys_pci_addr.bus)),
            bool(self.slot_regex.match(phys_pci_addr.slot)),
            bool(self.func_regex.match(phys_pci_addr.func))
            ]
        return all(conditions)


class WhitelistPciAddress(object):
    """Manages the address fields of the whitelist.

    This class checks the address fields of the pci.passthrough_whitelist
    configuration option, validating the address fields.
    Example configs:

        | [pci]
        | passthrough_whitelist = {"address":"*:0a:00.*",
        |                          "physical_network":"physnet1"}
        | passthrough_whitelist = {"address": {"domain": ".*",
                                               "bus": "02",
                                               "slot": "01",
                                               "function": "[0-2]"},
                                    "physical_network":"net1"}
        | passthrough_whitelist = {"vendor_id":"1137","product_id":"0071"}

    """
    def __init__(self, pci_addr, is_physical_function):
        self.is_physical_function = is_physical_function
        self._init_address_fields(pci_addr)

    def _check_physical_function(self):
        if self.pci_address_spec.is_single_address():
            self.is_physical_function = (
                utils.is_physical_function(
                    self.pci_address_spec.domain,
                    self.pci_address_spec.bus,
                    self.pci_address_spec.slot,
                    self.pci_address_spec.func))

    def _init_address_fields(self, pci_addr):
        if not self.is_physical_function:
            if isinstance(pci_addr, six.string_types):
                self.pci_address_spec = PciAddressGlobSpec(pci_addr)
            elif isinstance(pci_addr, dict):
                self.pci_address_spec = PciAddressRegexSpec(pci_addr)
            else:
                raise exception.PciDeviceWrongAddressFormat(address=pci_addr)
            self._check_physical_function()
        else:
            self.pci_address_spec = PhysicalPciAddress(pci_addr)

    def match(self, pci_addr, pci_phys_addr):
        """Match a device to this PciAddress.  Assume this is called given
        pci_addr and pci_phys_addr reported by libvirt, no attempt is made to
        verify if pci_addr is a VF of pci_phys_addr.

        :param pci_addr: PCI address of the device to match.
        :param pci_phys_addr: PCI address of the parent of the device to match
                              (or None if the device is not a VF).
        """

        # Try to match on the parent PCI address if the PciDeviceSpec is a
        # PF (sriov is available) and the device to match is a VF.  This
        # makes it possible to specify the PCI address of a PF in the
        # pci.passthrough_whitelist to match any of its VFs' PCI addresses.
        if self.is_physical_function and pci_phys_addr:
            pci_phys_addr_obj = PhysicalPciAddress(pci_phys_addr)
            if self.pci_address_spec.match(pci_phys_addr_obj):
                return True

        # Try to match on the device PCI address only.
        pci_addr_obj = PhysicalPciAddress(pci_addr)
        return self.pci_address_spec.match(pci_addr_obj)


class PciDeviceSpec(PciAddressSpec):
    def __init__(self, dev_spec):
        self.tags = dev_spec
        self._init_dev_details()

    def _init_dev_details(self):
        self.vendor_id = self.tags.pop("vendor_id", ANY)
        self.product_id = self.tags.pop("product_id", ANY)
        # Note(moshele): The address attribute can be a string or a dict.
        # For glob syntax or specific pci it is a string and for regex syntax
        # it is a dict. The WhitelistPciAddress class handles both types.
        self.address = self.tags.pop("address", None)
        self.dev_name = self.tags.pop("devname", None)

        self.vendor_id = self.vendor_id.strip()
        self._set_pci_dev_info('vendor_id', MAX_VENDOR_ID, '%04x')
        self._set_pci_dev_info('product_id', MAX_PRODUCT_ID, '%04x')

        if self.address and self.dev_name:
            raise exception.PciDeviceInvalidDeviceName()
        if not self.dev_name:
            pci_address = self.address or "*:*:*.*"
            self.address = WhitelistPciAddress(pci_address, False)

    def match(self, dev_dict):
        if self.dev_name:
            address_str, pf = utils.get_function_by_ifname(
                self.dev_name)
            if not address_str:
                return False
            # Note(moshele): In this case we always passing a string
            # of the PF pci address
            address_obj = WhitelistPciAddress(address_str, pf)
        elif self.address:
            address_obj = self.address
        return all([
            self.vendor_id in (ANY, dev_dict['vendor_id']),
            self.product_id in (ANY, dev_dict['product_id']),
            address_obj.match(dev_dict['address'],
                dev_dict.get('parent_addr'))])

    def match_pci_obj(self, pci_obj):
        return self.match({'vendor_id': pci_obj.vendor_id,
                           'product_id': pci_obj.product_id,
                           'address': pci_obj.address,
                           'parent_addr': pci_obj.parent_addr})

    def get_tags(self):
        return self.tags



# Copyright 2015 Cloudbase Solutions SRL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Unit tests for the Hyper-V RDPConsoleOps.
"""

import mock

from nova.tests.unit.virt.hyperv import test_base
from nova.virt.hyperv import rdpconsoleops


class RDPConsoleOpsTestCase(test_base.HyperVBaseTestCase):

    def setUp(self):
        super(RDPConsoleOpsTestCase, self).setUp()

        self.rdpconsoleops = rdpconsoleops.RDPConsoleOps()
        self.rdpconsoleops._hostops = mock.MagicMock()
        self.rdpconsoleops._vmutils = mock.MagicMock()
        self.rdpconsoleops._rdpconsoleutils = mock.MagicMock()

    def test_get_rdp_console(self):
        mock_get_host_ip = self.rdpconsoleops._hostops.get_host_ip_addr
        mock_get_rdp_port = (
            self.rdpconsoleops._rdpconsoleutils.get_rdp_console_port)
        mock_get_vm_id = self.rdpconsoleops._vmutils.get_vm_id

        connect_info = self.rdpconsoleops.get_rdp_console(mock.DEFAULT)

        self.assertEqual(mock_get_host_ip.return_value, connect_info.host)
        self.assertEqual(mock_get_rdp_port.return_value, connect_info.port)
        self.assertEqual(mock_get_vm_id.return_value,
                         connect_info.internal_access_path)



# Copyright (c) 2015 Ericsson AB
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Affinity Weighers.  Weigh hosts by the number of instances from a given host.

AffinityWeigher implements the soft-affinity policy for server groups by
preferring the hosts that has more instances from the given group.

AntiAffinityWeigher implements the soft-anti-affinity policy for server groups
by preferring the hosts that has less instances from the given group.

"""
from oslo_config import cfg
from oslo_log import log as logging

from nova.scheduler import utils
from nova.scheduler import weights

CONF = cfg.CONF

LOG = logging.getLogger(__name__)


class _SoftAffinityWeigherBase(weights.BaseHostWeigher):
    policy_name = None

    def _weigh_object(self, host_state, request_spec):
        """Higher weights win."""
        if not request_spec.instance_group:
            return 0

        policy = request_spec.instance_group.policy

        if self.policy_name != policy:
            return 0

        instances = set(host_state.instances.keys())
        members = set(request_spec.instance_group.members)
        member_on_host = instances.intersection(members)

        return len(member_on_host)


class ServerGroupSoftAffinityWeigher(_SoftAffinityWeigherBase):
    policy_name = 'soft-affinity'

    def weight_multiplier(self, host_state):
        return utils.get_weight_multiplier(
            host_state, 'soft_affinity_weight_multiplier',
            CONF.filter_scheduler.soft_affinity_weight_multiplier)


class ServerGroupSoftAntiAffinityWeigher(_SoftAffinityWeigherBase):
    policy_name = 'soft-anti-affinity'

    def weight_multiplier(self, host_state):
        return utils.get_weight_multiplier(
            host_state, 'soft_anti_affinity_weight_multiplier',
            CONF.filter_scheduler.soft_anti_affinity_weight_multiplier)

    def _weigh_object(self, host_state, request_spec):
        weight = super(ServerGroupSoftAntiAffinityWeigher, self)._weigh_object(
            host_state, request_spec)
        return -1 * weight



# Copyright 2016 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import time

import nova.scheduler.utils
import nova.servicegroup
from nova import test
from nova.tests import fixtures as nova_fixtures
from nova.tests.functional.api import client
from nova.tests.unit import cast_as_call
import nova.tests.unit.image.fake
from nova.tests.unit import policy_fixture


class TestServerGet(test.TestCase):
    REQUIRES_LOCKING = True

    def setUp(self):
        super(TestServerGet, self).setUp()
        self.useFixture(policy_fixture.RealPolicyFixture())
        self.useFixture(nova_fixtures.NeutronFixture(self))
        api_fixture = self.useFixture(nova_fixtures.OSAPIFixture(
            api_version='v2.1'))

        # The non-admin API client is fine to stay at 2.1 since it just creates
        # and deletes the server.
        self.api = api_fixture.api
        self.admin_api = api_fixture.admin_api
        # The admin API client needs to be at microversion 2.16 to exhibit the
        # regression.
        self.admin_api.microversion = '2.16'

        # the image fake backend needed for image discovery
        nova.tests.unit.image.fake.stub_out_image_service(self)

        self.start_service('conductor')
        self.start_service('scheduler')
        self.compute = self.start_service('compute')

        self.useFixture(cast_as_call.CastAsCall(self))
        self.addCleanup(nova.tests.unit.image.fake.FakeImageService_reset)

        self.image_id = self.api.get_images()[0]['id']
        self.flavor_id = self.api.get_flavors()[0]['id']

    def test_list_deleted_instances(self):
        """Regression test for bug #1548980.

        Before fixing this bug, listing deleted instances returned a 404
        because lazy-loading services from a deleted instance failed. Now
        we should be able to list the deleted instance and the host_state
        attribute should be "".
        """
        server = dict(name='server1',
                      imageRef=self.image_id,
                      flavorRef=self.flavor_id)
        server = self.api.post_server({'server': server})
        self.api.delete_server(server['id'])
        # Wait 30 seconds for it to be gone.
        for x in range(30):
            try:
                self.api.get_server(server['id'])
                time.sleep(1)
            except client.OpenStackApiNotFoundException:
                break
        else:
            self.fail('Timed out waiting to delete server: %s' % server['id'])
        servers = self.admin_api.get_servers(search_opts={'deleted': 1})
        self.assertEqual(1, len(servers))
        self.assertEqual(server['id'], servers[0]['id'])
        # host_status is returned in the 2.16 microversion and since the server
        # is deleted it should be the empty string
        self.assertEqual(0, len(servers[0]['host_status']))



# Copyright (c) 2017 Veritas Technologies LLC
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from os_brick.initiator import connector

from nova.tests.unit.virt.libvirt.volume import test_volume
from nova.virt.libvirt.volume import vrtshyperscale

DEVICE_NAME = '{8ee71c33-dcd0-4267-8f2b-e0742ecabe9f}'
DEVICE_PATH = '/dev/8ee71c33-dcd0-4267-8f2b-e0742ec'


class LibvirtHyperScaleVolumeDriverTestCase(
        test_volume.LibvirtVolumeBaseTestCase):

    def test_driver_init(self):
        hs = vrtshyperscale.LibvirtHyperScaleVolumeDriver(self.fake_host)
        self.assertIsInstance(hs.connector, connector.HyperScaleConnector)

    def test_get_config(self):
        hs = vrtshyperscale.LibvirtHyperScaleVolumeDriver(self.fake_host)

        # expect valid conf is returned if called with proper arguments
        conn = {'data': {'device_path': DEVICE_PATH}}

        conf = hs.get_config(conn, self.disk_info)

        self.assertEqual("block", conf.source_type)
        self.assertEqual(DEVICE_PATH, conf.source_path)

    @mock.patch('os_brick.initiator.connectors.vrtshyperscale'
                '.HyperScaleConnector.connect_volume')
    def test_connect_volume(self, mock_brick_connect_volume):
        mock_brick_connect_volume.return_value = {'path': DEVICE_PATH}

        hs = vrtshyperscale.LibvirtHyperScaleVolumeDriver(self.fake_host)

        # dummy arguments are just passed through to mock connector
        disk_info = {'name': DEVICE_NAME}
        connection_info = {'data': disk_info}

        hs.connect_volume(connection_info, mock.sentinel.instance)

        # expect connect_volume to add device_path to connection_info:
        self.assertEqual(connection_info['data']['device_path'], DEVICE_PATH)

    @mock.patch('os_brick.initiator.connectors.vrtshyperscale'
                '.HyperScaleConnector.disconnect_volume')
    def test_disconnect_volume(self, mock_brick_disconnect_volume):
        mock_brick_disconnect_volume.return_value = None

        hs = vrtshyperscale.LibvirtHyperScaleVolumeDriver(self.fake_host)

        # dummy arguments are just passed through to mock connector
        conn_data = {'name': DEVICE_NAME}
        connection_info = {'data': conn_data}

        hs.disconnect_volume(connection_info, mock.sentinel.instance)

        hs.connector.disconnect_volume.assert_called_once_with(conn_data, None)



# Copyright 2013 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from nova.tests.functional.api_sample_tests import test_servers


class AvailabilityZoneJsonTest(test_servers.ServersSampleBase):
    ADMIN_API = True
    sample_dir = "os-availability-zone"

    # Do not use the AvailabilityZoneFixture in the base class.
    # TODO(mriedem): Make this more realistic by creating a "us-west" zone
    # and putting the "compute" service host in it.
    availability_zones = []

    def test_availability_zone_list(self):
        response = self._do_get('os-availability-zone')
        self._verify_response('availability-zone-list-resp', {}, response, 200)

    def test_availability_zone_detail(self):
        response = self._do_get('os-availability-zone/detail')
        self._verify_response('availability-zone-detail-resp', {}, response,
                              200)



# Copyright 2016 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from keystoneauth1 import loading as ks_loading
from oslo_config import cfg

from nova.conf import utils as confutils


DEFAULT_SERVICE_TYPE = 'network'
NEUTRON_GROUP = 'neutron'

neutron_group = cfg.OptGroup(
    NEUTRON_GROUP,
    title='Neutron Options',
    help="""
Configuration options for neutron (network connectivity as a service).
""")

neutron_opts = [
    cfg.StrOpt('ovs_bridge',
         default='br-int',
         help="""
Default name for the Open vSwitch integration bridge.

Specifies the name of an integration bridge interface used by OpenvSwitch.
This option is only used if Neutron does not specify the OVS bridge name in
port binding responses.
"""),
    cfg.StrOpt('default_floating_pool',
        default='nova',
        help="""
Default name for the floating IP pool.

Specifies the name of floating IP pool used for allocating floating IPs. This
option is only used if Neutron does not specify the floating IP pool name in
port binding reponses.
"""),
    cfg.IntOpt('extension_sync_interval',
         default=600,
         min=0,
         help="""
Integer value representing the number of seconds to wait before querying
Neutron for extensions.  After this number of seconds the next time Nova
needs to create a resource in Neutron it will requery Neutron for the
extensions that it has loaded.  Setting value to 0 will refresh the
extensions with no wait.
"""),
    cfg.ListOpt('physnets',
        default=[],
        help="""
List of physnets present on this host.

For each *physnet* listed, an additional section,
``[neutron_physnet_$PHYSNET]``, will be added to the configuration file. Each
section must be configured with a single configuration option, ``numa_nodes``,
which should be a list of node IDs for all NUMA nodes this physnet is
associated with. For example::

    [neutron]
    physnets = foo, bar

    [neutron_physnet_foo]
    numa_nodes = 0

    [neutron_physnet_bar]
    numa_nodes = 0,1

Any *physnet* that is not listed using this option will be treated as having no
particular NUMA node affinity.

Tunnelled networks (VXLAN, GRE, ...) cannot be accounted for in this way and
are instead configured using the ``[neutron_tunnel]`` group. For example::

    [neutron_tunnel]
    numa_nodes = 1

Related options:

* ``[neutron_tunnel] numa_nodes`` can be used to configure NUMA affinity for
  all tunneled networks
* ``[neutron_physnet_$PHYSNET] numa_nodes`` must be configured for each value
  of ``$PHYSNET`` specified by this option
"""),
    cfg.IntOpt('http_retries',
               default=3,
               min=0,
               help="""
Number of times neutronclient should retry on any failed http call.

0 means connection is attempted only once. Setting it to any positive integer
means that on failure connection is retried that many times e.g. setting it
to 3 means total attempts to connect will be 4.

Possible values:

* Any integer value. 0 means connection is attempted only once
"""),
]

metadata_proxy_opts = [
    cfg.BoolOpt("service_metadata_proxy",
        default=False,
        help="""
When set to True, this option indicates that Neutron will be used to proxy
metadata requests and resolve instance ids. Otherwise, the instance ID must be
passed to the metadata request in the 'X-Instance-ID' header.

Related options:

* metadata_proxy_shared_secret
"""),
     cfg.StrOpt("metadata_proxy_shared_secret",
        default="",
        secret=True,
        help="""
This option holds the shared secret string used to validate proxy requests to
Neutron metadata requests. In order to be used, the
'X-Metadata-Provider-Signature' header must be supplied in the request.

Related options:

* service_metadata_proxy
"""),
]

ALL_OPTS = (neutron_opts + metadata_proxy_opts)


def register_opts(conf):
    conf.register_group(neutron_group)
    conf.register_opts(ALL_OPTS, group=neutron_group)
    confutils.register_ksa_opts(conf, neutron_group, DEFAULT_SERVICE_TYPE)


def register_dynamic_opts(conf):
    """Register dynamically-generated options and groups.

    This must be called by the service that wishes to use the options **after**
    the initial configuration has been loaded.
    """
    opt = cfg.ListOpt('numa_nodes', default=[], item_type=cfg.types.Integer())

    # Register the '[neutron_tunnel] numa_nodes' opt, implicitly
    # registering the '[neutron_tunnel]' group in the process. This could
    # be done statically but is done to avoid this group appearing in
    # nova.conf documentation while the other group does not.
    conf.register_opt(opt, group='neutron_tunnel')

    # Register the '[neutron_physnet_$PHYSNET] numa_nodes' opts, implicitly
    # registering the '[neutron_physnet_$PHYSNET]' groups in the process
    for physnet in conf.neutron.physnets:
        conf.register_opt(opt, group='neutron_physnet_%s' % physnet)


def list_opts():
    return {
        neutron_group: (
            ALL_OPTS +
            ks_loading.get_session_conf_options() +
            ks_loading.get_auth_common_conf_options() +
            ks_loading.get_auth_plugin_conf_options('password') +
            ks_loading.get_auth_plugin_conf_options('v2password') +
            ks_loading.get_auth_plugin_conf_options('v3password') +
            confutils.get_ksa_adapter_opts(DEFAULT_SERVICE_TYPE))
    }



# Copyright 2013 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import six
import webob

from nova.api.openstack.compute import suspend_server as \
    suspend_server_v21
from nova import objects
from nova.tests.unit.api.openstack.compute import admin_only_action_common
from nova.tests.unit.api.openstack import fakes


class SuspendServerTestsV21(admin_only_action_common.CommonTests):
    suspend_server = suspend_server_v21
    controller_name = 'SuspendServerController'
    _api_version = '2.1'

    def setUp(self):
        super(SuspendServerTestsV21, self).setUp()
        self.controller = getattr(self.suspend_server, self.controller_name)()
        self.compute_api = self.controller.compute_api
        self.stub_out('nova.api.openstack.compute.suspend_server.'
                      'SuspendServerController',
                      lambda *a, **kw: self.controller)

    def test_suspend_resume(self):
        self._test_actions(['_suspend', '_resume'])

    @mock.patch('nova.virt.hardware.get_mem_encryption_constraint',
                new=mock.Mock(return_value=True))
    @mock.patch.object(objects.instance.Instance, 'image_meta')
    def test_suspend_sev_rejected(self, mock_image):
        instance = self._stub_instance_get()
        ex = self.assertRaises(webob.exc.HTTPConflict,
                               self.controller._suspend,
                               self.req, fakes.FAKE_UUID, body={})
        self.assertIn("Operation 'suspend' not supported for SEV-enabled "
                      "instance (%s)" % instance.uuid, six.text_type(ex))

    def test_suspend_resume_with_non_existed_instance(self):
        self._test_actions_with_non_existed_instance(['_suspend', '_resume'])

    def test_suspend_resume_raise_conflict_on_invalid_state(self):
        self._test_actions_raise_conflict_on_invalid_state(['_suspend',
                                                            '_resume'])

    def test_actions_with_locked_instance(self):
        self._test_actions_with_locked_instance(['_suspend', '_resume'])



# Copyright 2014 NEC Corporation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from nova.api.validation import parameter_types

common_quota = {
    'type': ['integer', 'string'],
    'pattern': '^-?[0-9]+$',
    # -1 is a flag value for unlimited
    'minimum': -1,
    # maximum's value is limited to db constant's MAX_INT
    # (in nova/db/constants.py)
    'maximum': 0x7FFFFFFF
}

quota_resources = {
    'instances': common_quota,
    'cores': common_quota,
    'ram': common_quota,
    'floating_ips': common_quota,
    'fixed_ips': common_quota,
    'metadata_items': common_quota,
    'key_pairs': common_quota,
    'security_groups': common_quota,
    'security_group_rules': common_quota,
    'injected_files': common_quota,
    'injected_file_content_bytes': common_quota,
    'injected_file_path_bytes': common_quota,
    'server_groups': common_quota,
    'server_group_members': common_quota,
    # NOTE(stephenfin): This will always be rejected since it was nova-network
    # only, but we need to allow users to submit it at a minimum
    'networks': common_quota
}

update_quota_set = copy.deepcopy(quota_resources)
update_quota_set.update({'force': parameter_types.boolean})

update_quota_set_v236 = copy.deepcopy(update_quota_set)
del update_quota_set_v236['fixed_ips']
del update_quota_set_v236['floating_ips']
del update_quota_set_v236['security_groups']
del update_quota_set_v236['security_group_rules']
del update_quota_set_v236['networks']

update = {
    'type': 'object',
    'properties': {
        'type': 'object',
        'quota_set': {
            'properties': update_quota_set,
            'additionalProperties': False,
        },
    },
    'required': ['quota_set'],
    'additionalProperties': False,
}

update_v236 = copy.deepcopy(update)
update_v236['properties']['quota_set']['properties'] = update_quota_set_v236

# 2.57 builds on 2.36 and removes injected_file* quotas.
update_quota_set_v257 = copy.deepcopy(update_quota_set_v236)
del update_quota_set_v257['injected_files']
del update_quota_set_v257['injected_file_content_bytes']
del update_quota_set_v257['injected_file_path_bytes']
update_v257 = copy.deepcopy(update_v236)
update_v257['properties']['quota_set']['properties'] = update_quota_set_v257

query_schema = {
    'type': 'object',
    'properties': {
        'user_id': parameter_types.multi_params({'type': 'string'})
    },
    # NOTE(gmann): This is kept True to keep backward compatibility.
    # As of now Schema validation stripped out the additional parameters and
    # does not raise 400. In microversion 2.75, we have blocked the additional
    # parameters.
    'additionalProperties': True
}

query_schema_275 = copy.deepcopy(query_schema)
query_schema_275['additionalProperties'] = False



# Copyright 2016 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_policy import policy

from nova.policies import base


BASE_POLICY_NAME = 'os_compute_api:os-extended-server-attributes'


extended_server_attributes_policies = [
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME,
        check_str=base.SYSTEM_ADMIN,
        description="""Return extended attributes for server.

This rule will control the visibility for a set of servers attributes:

- ``OS-EXT-SRV-ATTR:host``
- ``OS-EXT-SRV-ATTR:instance_name``
- ``OS-EXT-SRV-ATTR:reservation_id`` (since microversion 2.3)
- ``OS-EXT-SRV-ATTR:launch_index`` (since microversion 2.3)
- ``OS-EXT-SRV-ATTR:hostname`` (since microversion 2.3)
- ``OS-EXT-SRV-ATTR:kernel_id`` (since microversion 2.3)
- ``OS-EXT-SRV-ATTR:ramdisk_id`` (since microversion 2.3)
- ``OS-EXT-SRV-ATTR:root_device_name`` (since microversion 2.3)
- ``OS-EXT-SRV-ATTR:user_data`` (since microversion 2.3)

Microvision 2.75 added the above attributes in the ``PUT /servers/{server_id}``
and ``POST /servers/{server_id}/action (rebuild)`` API responses which are
also controlled by this policy rule, like the ``GET /servers*`` APIs.
""",
        operations=[
            {
                'method': 'GET',
                'path': '/servers/{id}'
            },
            {
                'method': 'GET',
                'path': '/servers/detail'
            },
            {
                'method': 'PUT',
                'path': '/servers/{server_id}'
            },
            {
                'method': 'POST',
                'path': '/servers/{server_id}/action (rebuild)'
            }
        ],
        scope_types=['system', 'project']
    ),
]


def list_rules():
    return extended_server_attributes_policies



#    Copyright 2011 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections

from cinderclient.v3 import client as cinder_client_v3
import mock
from requests_mock.contrib import fixture

import nova.conf
from nova import context
from nova import exception
from nova import test
from nova.volume import cinder


CONF = nova.conf.CONF

_image_metadata = {
    'kernel_id': 'fake',
    'ramdisk_id': 'fake'
}


_volume_id = "6edbc2f4-1507-44f8-ac0d-eed1d2608d38"
_instance_uuid = "f4fda93b-06e0-4743-8117-bc8bcecd651b"
_instance_uuid_2 = "f4fda93b-06e0-4743-8117-bc8bcecd651c"
_attachment_id = "3b4db356-253d-4fab-bfa0-e3626c0b8405"
_attachment_id_2 = "3b4db356-253d-4fab-bfa0-e3626c0b8406"
_device = "/dev/vdb"
_device_2 = "/dev/vdc"


_volume_attachment = \
    [{"server_id": _instance_uuid,
      "attachment_id": _attachment_id,
      "host_name": "",
      "volume_id": _volume_id,
      "device": _device,
      "id": _volume_id
    }]


_volume_attachment_2 = _volume_attachment
_volume_attachment_2.append({"server_id": _instance_uuid_2,
                             "attachment_id": _attachment_id_2,
                             "host_name": "",
                             "volume_id": _volume_id,
                             "device": _device_2,
                             "id": _volume_id})


exp_volume_attachment = collections.OrderedDict()
exp_volume_attachment[_instance_uuid] = {'attachment_id': _attachment_id,
                                         'mountpoint': _device}
exp_volume_attachment_2 = exp_volume_attachment
exp_volume_attachment_2[_instance_uuid_2] = {'attachment_id': _attachment_id_2,
                                             'mountpoint': _device_2}


class BaseCinderTestCase(object):

    def setUp(self):
        super(BaseCinderTestCase, self).setUp()
        cinder.reset_globals()
        self.requests = self.useFixture(fixture.Fixture())
        self.api = cinder.API()

        self.context = context.RequestContext('username',
                                              'project_id',
                                              auth_token='token',
                                              service_catalog=self.CATALOG)

    def flags(self, *args, **kwargs):
        super(BaseCinderTestCase, self).flags(*args, **kwargs)
        cinder.reset_globals()

    def create_client(self):
        return cinder.cinderclient(self.context)

    def test_context_with_catalog(self):
        self.assertEqual(self.URL, self.create_client().client.get_endpoint())

    def test_cinder_http_retries(self):
        retries = 42
        self.flags(http_retries=retries, group='cinder')
        self.assertEqual(retries, self.create_client().client.connect_retries)

    def test_cinder_api_insecure(self):
        # The True/False negation is awkward, but better for the client
        # to pass us insecure=True and we check verify_cert == False
        self.flags(insecure=True, group='cinder')
        self.assertFalse(self.create_client().client.session.verify)

    def test_cinder_http_timeout(self):
        timeout = 123
        self.flags(timeout=timeout, group='cinder')
        self.assertEqual(timeout, self.create_client().client.session.timeout)

    def test_cinder_api_cacert_file(self):
        cacert = "/etc/ssl/certs/ca-certificates.crt"
        self.flags(cafile=cacert, group='cinder')
        self.assertEqual(cacert, self.create_client().client.session.verify)


# NOTE(mriedem): This does not extend BaseCinderTestCase because Cinder v1 is
# no longer supported, this is just to test that trying to use v1 fails.
class CinderV1TestCase(test.NoDBTestCase):

    @mock.patch('nova.volume.cinder.cinder_client.get_volume_api_from_url',
                return_value='1')
    def test_cinderclient_unsupported_v1(self, get_api_version):
        """Tests that we fail if trying to use Cinder v1."""
        self.flags(catalog_info='volume:cinder:publicURL', group='cinder')
        # setup mocks
        get_endpoint = mock.Mock(
            return_value='http://localhost:8776/v1/%(project_id)s')
        fake_session = mock.Mock(get_endpoint=get_endpoint)
        ctxt = context.get_context()
        with mock.patch.object(cinder, '_SESSION', fake_session):
            self.assertRaises(exception.UnsupportedCinderAPIVersion,
                              cinder.cinderclient, ctxt)
        get_api_version.assert_called_once_with(get_endpoint.return_value)


# NOTE(mriedem): This does not extend BaseCinderTestCase because Cinder v2 is
# no longer supported, this is just to test that trying to use v2 fails.
class CinderV2TestCase(test.NoDBTestCase):

    @mock.patch('nova.volume.cinder.cinder_client.get_volume_api_from_url',
                return_value='2')
    def test_cinderclient_unsupported_v2(self, get_api_version):
        """Tests that we fail if trying to use Cinder v2."""
        self.flags(catalog_info='volumev2:cinderv2:publicURL', group='cinder')
        # setup mocks
        get_endpoint = mock.Mock(
            return_value='http://localhost:8776/v2/%(project_id)s')
        fake_session = mock.Mock(get_endpoint=get_endpoint)
        ctxt = context.get_context()
        with mock.patch.object(cinder, '_SESSION', fake_session):
            self.assertRaises(exception.UnsupportedCinderAPIVersion,
                              cinder.cinderclient, ctxt)
        get_api_version.assert_called_once_with(get_endpoint.return_value)


class CinderV3TestCase(BaseCinderTestCase, test.NoDBTestCase):
    """Test case for cinder volume v3 api."""

    URL = "http://localhost:8776/v3/project_id"

    CATALOG = [{
        "type": "volumev3",
        "name": "cinderv3",
        "endpoints": [{"publicURL": URL}]
    }]

    def setUp(self):
        super(CinderV3TestCase, self).setUp()
        self.addCleanup(CONF.reset)

    def create_client(self):
        c = super(CinderV3TestCase, self).create_client()
        self.assertIsInstance(c, cinder_client_v3.Client)
        self.assertEqual('3.0', c.api_version.get_string())
        return c

    def stub_volume(self, **kwargs):
        volume = {
            'name': None,
            'description': None,
            "attachments": [],
            "availability_zone": "cinderv2",
            "created_at": "2013-08-10T00:00:00.000000",
            "id": _volume_id,
            "metadata": {},
            "size": 1,
            "snapshot_id": None,
            "status": "available",
            "volume_type": "None",
            "bootable": "true",
            "multiattach": "true"
        }
        volume.update(kwargs)
        return volume

    def test_cinder_endpoint_template(self):
        endpoint = 'http://other_host:8776/v3/%(project_id)s'
        self.flags(endpoint_template=endpoint, group='cinder')
        self.assertEqual('http://other_host:8776/v3/project_id',
                         self.create_client().client.endpoint_override)

    def test_get_non_existing_volume(self):
        self.requests.get(self.URL + '/volumes/nonexisting',
                          status_code=404)

        self.assertRaises(exception.VolumeNotFound, self.api.get, self.context,
                          'nonexisting')

    def test_volume_with_image_metadata(self):
        v = self.stub_volume(id='1234', volume_image_metadata=_image_metadata)
        self.requests.get(self.URL + '/volumes/5678', json={'volume': v})
        volume = self.api.get(self.context, '5678')
        self.assertIn('volume_image_metadata', volume)
        self.assertEqual(_image_metadata, volume['volume_image_metadata'])

    def test_volume_without_attachment(self):
        v = self.stub_volume(id='1234')
        self.requests.get(self.URL + '/volumes/5678', json={'volume': v})
        volume = self.api.get(self.context, '5678')
        self.assertIsNone(volume.get('attachments'))

    def test_volume_with_one_attachment(self):
        v = self.stub_volume(id='1234', attachments=_volume_attachment)
        self.requests.get(self.URL + '/volumes/5678', json={'volume': v})
        volume = self.api.get(self.context, '5678')
        self.assertIn('attachments', volume)
        self.assertEqual(exp_volume_attachment, volume['attachments'])

    def test_volume_with_two_attachments(self):
        v = self.stub_volume(id='1234', attachments=_volume_attachment_2)
        self.requests.get(self.URL + '/volumes/5678', json={'volume': v})
        volume = self.api.get(self.context, '5678')
        self.assertIn('attachments', volume)
        self.assertEqual(exp_volume_attachment_2, volume['attachments'])

    def test_create_client_with_no_service_name(self):
        """Tests that service_name is not required and not passed through
        when constructing the cinder client Client object if it's not
        configured.
        """
        self.flags(catalog_info='volumev3::public', group='cinder')
        with mock.patch('cinderclient.client.Client') as mock_client:
            # We don't use self.create_client() because that has additional
            # assertions that we don't care about in this test. We just care
            # about how the client is created, not what is returned.
            cinder.cinderclient(self.context)
        self.assertEqual(1, len(mock_client.call_args_list))
        call_kwargs = mock_client.call_args_list[0][1]
        # Make sure service_type and interface are passed through.
        self.assertEqual('volumev3', call_kwargs['service_type'])
        self.assertEqual('public', call_kwargs['interface'])
        # And service_name is not passed through.
        self.assertNotIn('service_name', call_kwargs)



# Copyright (c) 2014 Rackspace Hosting
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from sqlalchemy import Column
from sqlalchemy import MetaData
from sqlalchemy import Table
from sqlalchemy import Text


def upgrade(engine):
    meta = MetaData()
    meta.bind = engine

    # Drop the compute_node_stats table and add a 'stats' column to
    # compute_nodes directly.  The data itself is transient and doesn't
    # need to be copied over.
    table_names = ('compute_node_stats', 'shadow_compute_node_stats')
    for table_name in table_names:
        table = Table(table_name, meta, autoload=True)
        table.drop()

    # Add a new stats column to compute nodes
    table_names = ('compute_nodes', 'shadow_compute_nodes')
    for table_name in table_names:
        table = Table(table_name, meta, autoload=True)
        stats = Column('stats', Text, default='{}')
        table.create_column(stats)



# Copyright (c) 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from nova.scheduler import filters

LOG = logging.getLogger(__name__)


class RetryFilter(filters.BaseHostFilter):
    """Filter out nodes that have already been attempted for scheduling
    purposes
    """

    # NOTE(danms): This does not affect _where_ an instance lands, so not
    # related to rebuild.
    RUN_ON_REBUILD = False

    def __init__(self):
        super(RetryFilter, self).__init__()
        LOG.warning('The RetryFilter is deprecated since the 20.0.0 Train '
                    'release. Since the 17.0.0 (Queens) release, the '
                    'scheduler has provided alternate hosts for rescheduling '
                    'so the scheduler does not need to be called during a '
                    'reschedule which makes the RetryFilter useless.')

    def host_passes(self, host_state, spec_obj):
        """Skip nodes that have already been attempted."""
        retry = spec_obj.retry
        if not retry:
            return True

        # TODO(sbauza): Once the HostState is actually a ComputeNode, we could
        # easily get this one...
        host = [host_state.host, host_state.nodename]
        # TODO(sbauza)... and we wouldn't need to primitive the hosts into
        # lists
        hosts = [[cn.host, cn.hypervisor_hostname] for cn in retry.hosts]

        passes = host not in hosts

        if not passes:
            LOG.info("Host %(host)s fails.  Previously tried hosts: "
                     "%(hosts)s", {'host': host, 'hosts': hosts})

        # Host passes if it's not in the list of previously attempted hosts:
        return passes



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for the testing base code."""

import os.path
import tempfile
import uuid

import mock
from oslo_log import log as logging
import oslo_messaging as messaging
import six

import nova.conf
from nova import exception
from nova import rpc
from nova import test
from nova.tests import fixtures

LOG = logging.getLogger(__name__)

CONF = nova.conf.CONF


class IsolationTestCase(test.TestCase):
    """Ensure that things are cleaned up after failed tests.

    These tests don't really do much here, but if isolation fails a bunch
    of other tests should fail.

    """
    def test_service_isolation(self):
        self.useFixture(fixtures.ServiceFixture('compute'))

    def test_rpc_consumer_isolation(self):
        class NeverCalled(object):

            def __getattribute__(self, name):
                if name == 'target':
                    # oslo.messaging 5.31.0 explicitly looks for 'target'
                    # on the endpoint and checks it's type, so we can't avoid
                    # it here, just ignore it if that's the case.
                    return
                assert False, "I should never get called. name: %s" % name

        server = rpc.get_server(messaging.Target(topic='compute',
                                                 server=CONF.host),
                                endpoints=[NeverCalled()])
        server.start()


class JsonTestCase(test.NoDBTestCase):
    def test_compare_dict_string(self):
        expected = {
            "employees": [
                {"firstName": "Anna", "lastName": "Smith"},
                {"firstName": "John", "lastName": "Doe"},
                {"firstName": "Peter", "lastName": "Jones"}
            ],
            "locations": set(['Boston', 'Mumbai', 'Beijing', 'Perth'])
        }
        actual = """{
    "employees": [
        {
            "lastName": "Doe",
            "firstName": "John"
        },
        {
            "lastName": "Smith",
            "firstName": "Anna"
        },
        {
            "lastName": "Jones",
            "firstName": "Peter"
        }
    ],
    "locations": [
        "Perth",
        "Boston",
        "Mumbai",
        "Beijing"
    ]
}"""
        self.assertJsonEqual(expected, actual)

    def test_fail_on_list_length(self):
        expected = {
            'top': {
                'l1': {
                    'l2': ['a', 'b', 'c']
                }
            }
        }
        actual = {
            'top': {
                'l1': {
                    'l2': ['c', 'a', 'b', 'd']
                }
            }
        }
        try:
            self.assertJsonEqual(expected, actual)
        except Exception as e:
            # error reported is going to be a cryptic length failure
            # on the level2 structure.
            self.assertEqual(
                ("3 != 4: path: root.top.l1.l2. Different list items\n"
                 "expected=['a', 'b', 'c']\n"
                 "observed=['a', 'b', 'c', 'd']\n"
                 "difference=['d']"),
                e.difference)
            self.assertIn(
                "actual:\n{'top': {'l1': {'l2': ['c', 'a', 'b', 'd']}}}",
                six.text_type(e))
            self.assertIn(
                "expected:\n{'top': {'l1': {'l2': ['a', 'b', 'c']}}}",
                six.text_type(e))
        else:
            self.fail("This should have raised a mismatch exception")

    def test_fail_on_dict_length(self):
        expected = {
            'top': {
                'l1': {
                    'l2': {'a': 1, 'b': 2, 'c': 3}
                }
            }
        }
        actual = {
            'top': {
                'l1': {
                    'l2': {'a': 1, 'b': 2}
                }
            }
        }
        try:
            self.assertJsonEqual(expected, actual)
        except Exception as e:
            self.assertEqual(
                ("3 != 2: path: root.top.l1.l2. Different dict key sets\n"
                 "expected=['a', 'b', 'c']\n"
                 "observed=['a', 'b']\n"
                 "difference=['c']"),
                e.difference)
        else:
            self.fail("This should have raised a mismatch exception")

    def test_fail_on_dict_keys(self):
        expected = {
            'top': {
                'l1': {
                    'l2': {'a': 1, 'b': 2, 'c': 3}
                }
            }
        }
        actual = {
            'top': {
                'l1': {
                    'l2': {'a': 1, 'b': 2, 'd': 3}
                }
            }
        }
        try:
            self.assertJsonEqual(expected, actual)
        except Exception as e:
            self.assertIn(
                "path: root.top.l1.l2. Dict keys are not equal",
                e.difference)
        else:
            self.fail("This should have raised a mismatch exception")

    def test_fail_on_list_value(self):
        expected = {
            'top': {
                'l1': {
                    'l2': ['a', 'b', 'c']
                }
            }
        }
        actual = {
            'top': {
                'l1': {
                    'l2': ['c', 'a', 'd']
                }
            }
        }
        try:
            self.assertJsonEqual(expected, actual)
        except Exception as e:
            self.assertEqual(
                "'b' != 'c': path: root.top.l1.l2[1]",
                e.difference)
            self.assertIn(
                "actual:\n{'top': {'l1': {'l2': ['c', 'a', 'd']}}}",
                six.text_type(e))
            self.assertIn(
                "expected:\n{'top': {'l1': {'l2': ['a', 'b', 'c']}}}",
                six.text_type(e))
        else:
            self.fail("This should have raised a mismatch exception")

    def test_fail_on_dict_value(self):
        expected = {
            'top': {
                'l1': {
                    'l2': {'a': 1, 'b': 2, 'c': 3}
                }
            }
        }
        actual = {
            'top': {
                'l1': {
                    'l2': {'a': 1, 'b': 2, 'c': 4}
                }
            }
        }
        try:
            self.assertJsonEqual(expected, actual, 'test message')
        except Exception as e:
            self.assertEqual(
                "3 != 4: path: root.top.l1.l2.c", e.difference)
            self.assertIn("actual:\n{'top': {'l1': {'l2': {", six.text_type(e))
            self.assertIn(
                "expected:\n{'top': {'l1': {'l2': {", six.text_type(e))
            self.assertIn(
                "message: test message\n", six.text_type(e))
        else:
            self.fail("This should have raised a mismatch exception")

    def test_compare_scalars(self):
        with self.assertRaisesRegex(AssertionError, 'True != False'):
            self.assertJsonEqual(True, False)


class BadLogTestCase(test.NoDBTestCase):
    """Make sure a mis-formatted debug log will get caught."""

    def test_bad_debug_log(self):
        self.assertRaises(KeyError,
            LOG.debug, "this is a misformated %(log)s", {'nothing': 'nothing'})


class MatchTypeTestCase(test.NoDBTestCase):

    def test_match_type_simple(self):
        matcher = test.MatchType(dict)

        self.assertEqual(matcher, {})
        self.assertEqual(matcher, {"hello": "world"})
        self.assertEqual(matcher, {"hello": ["world"]})
        self.assertNotEqual(matcher, [])
        self.assertNotEqual(matcher, [{"hello": "world"}])
        self.assertNotEqual(matcher, 123)
        self.assertNotEqual(matcher, "foo")

    def test_match_type_object(self):
        class Hello(object):
            pass

        class World(object):
            pass

        matcher = test.MatchType(Hello)

        self.assertEqual(matcher, Hello())
        self.assertNotEqual(matcher, World())
        self.assertNotEqual(matcher, 123)
        self.assertNotEqual(matcher, "foo")


class ContainKeyValueTestCase(test.NoDBTestCase):

    def test_contain_key_value_normal(self):
        matcher = test.ContainKeyValue('foo', 'bar')

        self.assertEqual(matcher, {123: 'nova', 'foo': 'bar'})
        self.assertNotEqual(matcher, {'foo': 123})
        self.assertNotEqual(matcher, {})

    def test_contain_key_value_exception(self):
        matcher = test.ContainKeyValue('foo', 'bar')

        # Raise TypeError
        self.assertNotEqual(matcher, 123)
        self.assertNotEqual(matcher, 'foo')
        # Raise KeyError
        self.assertNotEqual(matcher, {1: 2, '3': 4, 5: '6'})
        self.assertNotEqual(matcher, {'bar': 'foo'})


class NovaExceptionReraiseFormatErrorTestCase(test.NoDBTestCase):
    """Test that format errors are reraised in tests."""
    def test_format_error_in_nova_exception(self):
        class FakeImageException(exception.NovaException):
            msg_fmt = 'Image %(image_id)s has wrong type %(type)s.'
        # wrong kwarg
        ex = self.assertRaises(KeyError, FakeImageException,
                               bogus='wrongkwarg')
        self.assertIn('image_id', six.text_type(ex))
        # no kwarg
        ex = self.assertRaises(KeyError, FakeImageException)
        self.assertIn('image_id', six.text_type(ex))
        # not enough kwargs
        ex = self.assertRaises(KeyError, FakeImageException, image_id='image')
        self.assertIn('type', six.text_type(ex))


class PatchExistsTestCase(test.NoDBTestCase):
    def test_with_patch_exists_true(self):
        """Test that "with patch_exists" can fake the existence of a file
        without changing other file existence checks, and that calls can
        be asserted on the mocked method.
        """
        self.assertFalse(os.path.exists('fake_file'))
        with self.patch_exists('fake_file', True) as mock_exists:
            self.assertTrue(os.path.exists('fake_file'))
            self.assertTrue(os.path.exists(__file__))
            self.assertFalse(os.path.exists('non-existent/file'))
            self.assertIn(mock.call('fake_file'), mock_exists.mock_calls)

    def test_with_patch_exists_false(self):
        """Test that "with patch_exists" can fake the non-existence of a file
        without changing other file existence checks, and that calls can
        be asserted on the mocked method.
        """
        self.assertTrue(os.path.exists(__file__))
        with self.patch_exists(__file__, False) as mock_exists:
            self.assertFalse(os.path.exists(__file__))
            self.assertTrue(os.path.exists(os.path.dirname(__file__)))
            self.assertFalse(os.path.exists('non-existent/file'))
            self.assertIn(mock.call(__file__), mock_exists.mock_calls)

    @test.patch_exists('fake_file', True)
    def test_patch_exists_decorator_true(self):
        """Test that @patch_exists can fake the existence of a file
        without changing other file existence checks.
        """
        self.assertTrue(os.path.exists('fake_file'))
        self.assertTrue(os.path.exists(__file__))
        self.assertFalse(os.path.exists('non-existent/file'))

    @test.patch_exists(__file__, False)
    def test_patch_exists_decorator_false(self):
        """Test that @patch_exists can fake the non-existence of a file
        without changing other file existence checks.
        """
        self.assertFalse(os.path.exists(__file__))
        self.assertTrue(os.path.exists(os.path.dirname(__file__)))
        self.assertFalse(os.path.exists('non-existent/file'))

    @test.patch_exists('fake_file1', True)
    @test.patch_exists('fake_file2', True)
    @test.patch_exists(__file__, False)
    def test_patch_exists_multiple_decorators(self):
        """Test that @patch_exists can be used multiple times on the
        same method.
        """
        self.assertTrue(os.path.exists('fake_file1'))
        self.assertTrue(os.path.exists('fake_file2'))
        self.assertFalse(os.path.exists(__file__))

        # Check non-patched parameters
        self.assertTrue(os.path.exists(os.path.dirname(__file__)))
        self.assertFalse(os.path.exists('non-existent/file'))


class PatchOpenTestCase(test.NoDBTestCase):
    fake_contents = "These file contents don't really exist"

    def _test_patched_open(self):
        """Test that a selectively patched open can fake the contents of a
        file while still allowing normal, real file operations.
        """
        self.assertFalse(os.path.exists('fake_file'))

        with open('fake_file') as f:
            self.assertEqual(self.fake_contents, f.read())

        # Test we can still open and read this file from within the
        # same context.  NOTE: We have to make sure we open the .py
        # file not the corresponding .pyc file.
        with open(__file__.rstrip('c')) as f:
            this_file_contents = f.read()
            self.assertIn("class %s(" % self.__class__.__name__,
                          this_file_contents)
            self.assertNotIn("magic concatenated" "string",
                             this_file_contents)

        # Test we can still create, write to, and then read from a
        # temporary file, from within the same context.
        tmp = tempfile.NamedTemporaryFile()
        tmp_contents = str(uuid.uuid1())
        with open(tmp.name, 'w') as f:
            f.write(tmp_contents)
        with open(tmp.name) as f:
            self.assertEqual(tmp_contents, f.read())

        return tmp.name

    def test_with_patch_open(self):
        """Test that "with patch_open" can fake the contents of a file
        without changing other file operations, and that calls can
        be asserted on the mocked method.
        """
        with self.patch_open('fake_file', self.fake_contents) as mock_open:
            tmp_name = self._test_patched_open()

            # Test we can make assertions about how the mock_open was called.
            self.assertIn(mock.call('fake_file'), mock_open.mock_calls)
            # The mock_open should get bypassed for non-patched path values:
            self.assertNotIn(mock.call(__file__), mock_open.mock_calls)
            self.assertNotIn(mock.call(tmp_name), mock_open.mock_calls)

    @test.patch_open('fake_file', fake_contents)
    def test_patch_open_decorator(self):
        """Test that @patch_open can fake the contents of a file
        without changing other file operations.
        """
        self._test_patched_open()



# Copyright 2012 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import webob

from nova.api.openstack import common
from nova.api.openstack.compute.schemas import remote_consoles
from nova.api.openstack import wsgi
from nova.api import validation
from nova.compute import api as compute
from nova import exception
from nova.policies import remote_consoles as rc_policies


class RemoteConsolesController(wsgi.Controller):
    def __init__(self):
        super(RemoteConsolesController, self).__init__()
        self.compute_api = compute.API()
        self.handlers = {'vnc': self.compute_api.get_vnc_console,
                         'spice': self.compute_api.get_spice_console,
                         'rdp': self.compute_api.get_rdp_console,
                         'serial': self.compute_api.get_serial_console,
                         'mks': self.compute_api.get_mks_console}

    @wsgi.Controller.api_version("2.1", "2.5")
    @wsgi.expected_errors((400, 404, 409, 501))
    @wsgi.action('os-getVNCConsole')
    @validation.schema(remote_consoles.get_vnc_console)
    def get_vnc_console(self, req, id, body):
        """Get text console output."""
        context = req.environ['nova.context']
        context.can(rc_policies.BASE_POLICY_NAME)

        # If type is not supplied or unknown, get_vnc_console below will cope
        console_type = body['os-getVNCConsole'].get('type')

        instance = common.get_instance(self.compute_api, context, id)
        try:
            output = self.compute_api.get_vnc_console(context,
                                                      instance,
                                                      console_type)
        except exception.ConsoleTypeUnavailable as e:
            raise webob.exc.HTTPBadRequest(explanation=e.format_message())
        except exception.InstanceNotFound as e:
            raise webob.exc.HTTPNotFound(explanation=e.format_message())
        except exception.InstanceNotReady as e:
            raise webob.exc.HTTPConflict(explanation=e.format_message())
        except NotImplementedError:
            common.raise_feature_not_supported()

        return {'console': {'type': console_type, 'url': output['url']}}

    @wsgi.Controller.api_version("2.1", "2.5")
    @wsgi.expected_errors((400, 404, 409, 501))
    @wsgi.action('os-getSPICEConsole')
    @validation.schema(remote_consoles.get_spice_console)
    def get_spice_console(self, req, id, body):
        """Get text console output."""
        context = req.environ['nova.context']
        context.can(rc_policies.BASE_POLICY_NAME)

        # If type is not supplied or unknown, get_spice_console below will cope
        console_type = body['os-getSPICEConsole'].get('type')

        instance = common.get_instance(self.compute_api, context, id)
        try:
            output = self.compute_api.get_spice_console(context,
                                                        instance,
                                                        console_type)
        except exception.ConsoleTypeUnavailable as e:
            raise webob.exc.HTTPBadRequest(explanation=e.format_message())
        except exception.InstanceNotFound as e:
            raise webob.exc.HTTPNotFound(explanation=e.format_message())
        except exception.InstanceNotReady as e:
            raise webob.exc.HTTPConflict(explanation=e.format_message())
        except NotImplementedError:
            common.raise_feature_not_supported()

        return {'console': {'type': console_type, 'url': output['url']}}

    @wsgi.Controller.api_version("2.1", "2.5")
    @wsgi.expected_errors((400, 404, 409, 501))
    @wsgi.action('os-getRDPConsole')
    @validation.schema(remote_consoles.get_rdp_console)
    def get_rdp_console(self, req, id, body):
        """Get text console output."""
        context = req.environ['nova.context']
        context.can(rc_policies.BASE_POLICY_NAME)

        # If type is not supplied or unknown, get_rdp_console below will cope
        console_type = body['os-getRDPConsole'].get('type')

        instance = common.get_instance(self.compute_api, context, id)
        try:
            # NOTE(mikal): get_rdp_console() can raise InstanceNotFound, so
            # we still need to catch it here.
            output = self.compute_api.get_rdp_console(context,
                                                      instance,
                                                      console_type)
        except exception.ConsoleTypeUnavailable as e:
            raise webob.exc.HTTPBadRequest(explanation=e.format_message())
        except exception.InstanceNotFound as e:
            raise webob.exc.HTTPNotFound(explanation=e.format_message())
        except exception.InstanceNotReady as e:
            raise webob.exc.HTTPConflict(explanation=e.format_message())
        except NotImplementedError:
            common.raise_feature_not_supported()

        return {'console': {'type': console_type, 'url': output['url']}}

    @wsgi.Controller.api_version("2.1", "2.5")
    @wsgi.expected_errors((400, 404, 409, 501))
    @wsgi.action('os-getSerialConsole')
    @validation.schema(remote_consoles.get_serial_console)
    def get_serial_console(self, req, id, body):
        """Get connection to a serial console."""
        context = req.environ['nova.context']
        context.can(rc_policies.BASE_POLICY_NAME)

        # If type is not supplied or unknown get_serial_console below will cope
        console_type = body['os-getSerialConsole'].get('type')
        instance = common.get_instance(self.compute_api, context, id)
        try:
            output = self.compute_api.get_serial_console(context,
                                                         instance,
                                                         console_type)
        except exception.InstanceNotFound as e:
            raise webob.exc.HTTPNotFound(explanation=e.format_message())
        except exception.InstanceNotReady as e:
            raise webob.exc.HTTPConflict(explanation=e.format_message())
        except (exception.ConsoleTypeUnavailable,
                exception.ImageSerialPortNumberInvalid,
                exception.ImageSerialPortNumberExceedFlavorValue,
                exception.SocketPortRangeExhaustedException) as e:
            raise webob.exc.HTTPBadRequest(explanation=e.format_message())
        except NotImplementedError:
            common.raise_feature_not_supported()

        return {'console': {'type': console_type, 'url': output['url']}}

    @wsgi.Controller.api_version("2.6")
    @wsgi.expected_errors((400, 404, 409, 501))
    @validation.schema(remote_consoles.create_v26, "2.6", "2.7")
    @validation.schema(remote_consoles.create_v28, "2.8")
    def create(self, req, server_id, body):
        context = req.environ['nova.context']
        instance = common.get_instance(self.compute_api, context, server_id)
        context.can(rc_policies.BASE_POLICY_NAME,
                    target={'project_id': instance.project_id})
        protocol = body['remote_console']['protocol']
        console_type = body['remote_console']['type']
        try:
            handler = self.handlers.get(protocol)
            output = handler(context, instance, console_type)
            return {'remote_console': {'protocol': protocol,
                                       'type': console_type,
                                       'url': output['url']}}

        except exception.InstanceNotFound as e:
            raise webob.exc.HTTPNotFound(explanation=e.format_message())
        except exception.InstanceNotReady as e:
            raise webob.exc.HTTPConflict(explanation=e.format_message())
        except (exception.ConsoleTypeInvalid,
                exception.ConsoleTypeUnavailable,
                exception.ImageSerialPortNumberInvalid,
                exception.ImageSerialPortNumberExceedFlavorValue,
                exception.SocketPortRangeExhaustedException) as e:
            raise webob.exc.HTTPBadRequest(explanation=e.format_message())
        except NotImplementedError:
            common.raise_feature_not_supported()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nova import test
from nova.tests import fixtures as nova_fixtures
from nova.tests.functional.api import client
from nova.tests.functional import integrated_helpers


class TestCreateServerGroupWithEmptyPolicies(
    test.TestCase, integrated_helpers.InstanceHelperMixin,
):
    """Demonstrate bug #1894966.

    Attempt to create a server group with an invalid 'policies' field. It
    should fail cleanly.
    """
    def setUp(self):
        super().setUp()

        api_fixture = self.useFixture(nova_fixtures.OSAPIFixture(
            api_version='v2.1'))
        self.api = api_fixture.api
        self.api.microversion = '2.63'  # the last version with the bug

    def test_create_with_empty_policies(self):
        exc = self.assertRaises(
            client.OpenStackApiException,
            self.api.post_server_groups,
            {'name': 'test group', 'policies': []})
        self.assertEqual(400, exc.response.status_code)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from nova import test
from nova.tests import fixtures as nova_fixtures
from nova.tests.functional import fixtures as func_fixtures
from nova.tests.functional import integrated_helpers
from nova.tests.unit import fake_network
from nova.tests.unit import fake_notifier
import nova.tests.unit.image.fake
from nova.tests.unit import policy_fixture


class TestEvacuationWithSourceReturningDuringRebuild(
        test.TestCase, integrated_helpers.InstanceHelperMixin):
    """Assert the behaviour of evacuating instances when the src returns early.

    This test asserts that evacuating instances end up in an ACTIVE state on
    the destination even when the source host comes back online during an
    evacuation while the migration record is in a pre-migrating state.
    """

    def setUp(self):
        super(TestEvacuationWithSourceReturningDuringRebuild, self).setUp()

        self.useFixture(policy_fixture.RealPolicyFixture())

        # The NeutronFixture is needed to stub out validate_networks in API.
        self.useFixture(nova_fixtures.NeutronFixture(self))

        # This stubs out the network allocation in compute.
        fake_network.set_stub_network_methods(self)

        # We need the computes reporting into placement for the filter
        # scheduler to pick a host.
        self.useFixture(func_fixtures.PlacementFixture())

        api_fixture = self.useFixture(nova_fixtures.OSAPIFixture(
            api_version='v2.1'))
        self.api = api_fixture.admin_api
        # 2.11 is needed for force_down
        # 2.14 is needed for evacuate without onSharedStorage flag
        self.api.microversion = '2.14'

        # the image fake backend needed for image discovery
        nova.tests.unit.image.fake.stub_out_image_service(self)
        self.addCleanup(nova.tests.unit.image.fake.FakeImageService_reset)

        self.start_service('conductor')
        self.start_service('scheduler')

        # Start two computes
        self._start_compute('host1')
        self._start_compute('host2')

        self.image_id = self.api.get_images()[0]['id']
        self.flavor_id = self.api.get_flavors()[0]['id']

        self.addCleanup(fake_notifier.reset)

        # Stub out rebuild with a slower method allowing the src compute to be
        # restarted once the migration hits pre-migrating after claiming
        # resources on the dest.
        manager_class = nova.compute.manager.ComputeManager
        original_rebuild = manager_class._do_rebuild_instance

        def start_src_rebuild(self_, context, instance, *args, **kwargs):
            server = self.api.get_server(instance.uuid)
            # Start the src compute once the migration is pre-migrating.
            self._wait_for_migration_status(server, ['pre-migrating'])
            self.computes.get(self.source_compute).start()
            original_rebuild(self_, context, instance, *args, **kwargs)

        self.stub_out('nova.compute.manager.ComputeManager.'
                      '_do_rebuild_instance', start_src_rebuild)

    def test_evacuation_with_source_compute_returning_during_rebuild(self):

        # Launch an instance
        server_request = {'name': 'server',
                          'imageRef': self.image_id,
                          'flavorRef': self.flavor_id}
        server_response = self.api.post_server({'server': server_request})
        server = self._wait_for_state_change(server_response,
                                             'ACTIVE')

        # Record where the instance is running before forcing the service down
        self.source_compute = server['OS-EXT-SRV-ATTR:host']
        self.computes.get(self.source_compute).stop()
        self.api.force_down_service(self.source_compute, 'nova-compute', True)

        # Start evacuating the instance from the source_host
        self.api.post_server_action(server['id'], {'evacuate': {}})

        # Wait for the instance to go into an ACTIVE state
        self._wait_for_state_change(server, 'ACTIVE')
        server = self.api.get_server(server['id'])
        host = server['OS-EXT-SRV-ATTR:host']
        migrations = self.api.get_migrations()

        # Assert that we have a single `done` migration record after the evac
        self.assertEqual(1, len(migrations))
        self.assertEqual('done', migrations[0]['status'])

        # Assert that the instance is now on the dest
        self.assertNotEqual(self.source_compute, host)



# Copyright 2014 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

import mock
from oslo_config import cfg
from oslo_utils.fixture import uuidsentinel as uuids
import webob

from nova.api.openstack.compute import tenant_networks \
        as networks_v21
from nova import exception
from nova import test
from nova.tests.unit.api.openstack import fakes

CONF = cfg.CONF

NETWORKS = [
    {
        "id": uuids.fake_net1,
        "name": "fake_net1",
    },
    {
        "id": uuids.fake_net2,
        "name": "fake_net2",
    }
]

DEFAULT_NETWORK = [
    {
        "id": uuids.fake_net3,
        "name": "default",
    }
]

NETWORKS_WITH_DEFAULT_NET = copy.deepcopy(NETWORKS)
NETWORKS_WITH_DEFAULT_NET.extend(DEFAULT_NETWORK)

DEFAULT_TENANT_ID = CONF.api.neutron_default_tenant_id


def fake_network_api_get_all(context):
    if context.project_id == DEFAULT_TENANT_ID:
        return DEFAULT_NETWORK
    else:
        return NETWORKS


class TenantNetworksTestV21(test.NoDBTestCase):
    ctrlr = networks_v21.TenantNetworkController
    validation_error = exception.ValidationError

    def setUp(self):
        # TODO(stephenfin): We should probably use NeutronFixture here
        super(TenantNetworksTestV21, self).setUp()
        # os-tenant-networks only supports Neutron when listing networks or
        # showing details about a network, create and delete operations
        # result in a 503 and 500 response, respectively.
        self.controller = self.ctrlr()
        self.req = fakes.HTTPRequest.blank('')
        self.original_value = CONF.api.use_neutron_default_nets

    def tearDown(self):
        super(TenantNetworksTestV21, self).tearDown()
        CONF.set_override("use_neutron_default_nets", self.original_value,
                          group='api')

    def test_network_show(self):
        with mock.patch.object(self.controller.network_api, 'get',
                               return_value=NETWORKS[0]):
            res = self.controller.show(self.req, 1)

        expected = {
            'id': NETWORKS[0]['id'],
            'label': NETWORKS[0]['name'],
            'cidr': str(None),
        }
        self.assertEqual(expected, res['network'])

    def test_network_show_not_found(self):
        ctxt = self.req.environ['nova.context']
        with mock.patch.object(self.controller.network_api, 'get',
                               side_effect=exception.NetworkNotFound(
                                   network_id=1)) as get_mock:
            self.assertRaises(webob.exc.HTTPNotFound,
                              self.controller.show, self.req, 1)
        get_mock.assert_called_once_with(ctxt, 1)

    def _test_network_index(self, default_net=True):
        CONF.set_override("use_neutron_default_nets", default_net, group='api')

        networks = NETWORKS
        if default_net:
            networks = NETWORKS_WITH_DEFAULT_NET

        expected = []
        for network in networks:
            expected.append({
                'id': network['id'],
                'label': network['name'],
                'cidr': str(None),
            })

        with mock.patch.object(self.controller.network_api, 'get_all',
                               side_effect=fake_network_api_get_all):
            res = self.controller.index(self.req)
        self.assertEqual(expected, res['networks'])

    def test_network_index_with_default_net(self):
        self._test_network_index()

    def test_network_index_without_default_net(self):
        self._test_network_index(default_net=False)


class TenantNetworksEnforcementV21(test.NoDBTestCase):

    def setUp(self):
        super(TenantNetworksEnforcementV21, self).setUp()
        self.controller = networks_v21.TenantNetworkController()
        self.req = fakes.HTTPRequest.blank('')

    def test_index_policy_failed(self):
        rule_name = 'os_compute_api:os-tenant-networks'
        self.policy.set_rules({rule_name: "project:non_fake"})
        exc = self.assertRaises(
            exception.PolicyNotAuthorized,
            self.controller.index,
            self.req)
        self.assertEqual(
            "Policy doesn't allow %s to be performed." % rule_name,
            exc.format_message())

    def test_show_policy_failed(self):
        rule_name = 'os_compute_api:os-tenant-networks'
        self.policy.set_rules({rule_name: "project:non_fake"})
        exc = self.assertRaises(
            exception.PolicyNotAuthorized,
            self.controller.show,
            self.req, fakes.FAKE_UUID)
        self.assertEqual(
            "Policy doesn't allow %s to be performed." % rule_name,
            exc.format_message())


class TenantNetworksDeprecationTest(test.NoDBTestCase):

    def setUp(self):
        super(TenantNetworksDeprecationTest, self).setUp()
        self.controller = networks_v21.TenantNetworkController()
        self.req = fakes.HTTPRequest.blank('', version='2.36')

    def test_all_apis_return_not_found(self):
        self.assertRaises(exception.VersionNotFoundForAPIMethod,
            self.controller.index, self.req)
        self.assertRaises(exception.VersionNotFoundForAPIMethod,
            self.controller.show, self.req, fakes.FAKE_UUID)



# Copyright 2020 Red Hat, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Validators for (preferrably) ``capabilities`` namespaced extra specs.

These are used by the ``ComputeCapabilitiesFilter`` scheduler filter. Note that
we explicitly do not allow the unnamespaced variant of extra specs since this
has been deprecated since Grizzly (commit 8ce8e4b6c0d). Users that insist on
using these can disable extra spec validation.

For all extra specs, the value can be one of the following:

* ``=`` (equal to or greater than as a number; same as vcpus case)
* ``==`` (equal to as a number)
* ``!=`` (not equal to as a number)
* ``>=`` (greater than or equal to as a number)
* ``<=`` (less than or equal to as a number)
* ``s==`` (equal to as a string)
* ``s!=`` (not equal to as a string)
* ``s>=`` (greater than or equal to as a string)
* ``s>`` (greater than as a string)
* ``s<=`` (less than or equal to as a string)
* ``s<`` (less than as a string)
* ``<in>`` (substring)
* ``<all-in>`` (all elements contained in collection)
* ``<or>`` (find one of these)
* A specific value, e.g. ``true``, ``123``, ``testing``

Examples are: ``>= 5``, ``s== 2.1.0``, ``<in> gcc``, ``<all-in> aes mmx``, and
``<or> fpu <or> gpu``
"""

from nova.api.validation.extra_specs import base


DESCRIPTION = """\
Specify that the '{capability}' capability provided by the host compute service
satisfy the provided filter value. Requires the ``ComputeCapabilitiesFilter``
scheduler filter.
"""

EXTRA_SPEC_VALIDATORS = []

# non-nested capabilities (from 'nova.objects.compute_node.ComputeNode' and
# nova.scheduler.host_manager.HostState')

for capability in (
    'id', 'uuid', 'service_id', 'host', 'vcpus', 'memory_mb', 'local_gb',
    'vcpus_used', 'memory_mb_used', 'local_gb_used',
    'hypervisor_type', 'hypervisor_version', 'hypervisor_hostname',
    'free_ram_mb', 'free_disk_gb', 'current_workload', 'running_vms',
    'disk_available_least', 'host_ip', 'mapped',
    'cpu_allocation_ratio', 'ram_allocation_ratio', 'disk_allocation_ratio',
) + (
    'total_usable_ram_mb', 'total_usable_disk_gb', 'disk_mb_used',
    'free_disk_mb', 'vcpus_total', 'vcpus_used', 'num_instances',
    'num_io_ops', 'failed_builds', 'aggregates', 'cell_uuid', 'updated',
):
    EXTRA_SPEC_VALIDATORS.append(
        base.ExtraSpecValidator(
            name=f'capabilities:{capability}',
            description=DESCRIPTION.format(capability=capability),
            value={
                # this is totally arbitary, since we need to support specific
                # values
                'type': str,
            },
        ),
    )


# nested capabilities (from 'nova.objects.compute_node.ComputeNode' and
# nova.scheduler.host_manager.HostState')

for capability in (
    'cpu_info', 'metrics', 'stats', 'numa_topology', 'supported_hv_specs',
    'pci_device_pools',
) + (
    'nodename', 'pci_stats', 'supported_instances', 'limits', 'instances',
):
    EXTRA_SPEC_VALIDATORS.extend([
        base.ExtraSpecValidator(
            name=f'capabilities:{capability}{{filter}}',
            description=DESCRIPTION.format(capability=capability),
            parameters=[
                {
                    'name': 'filter',
                    # this is optional, but if it's present it must be preceded
                    # by ':'
                    'pattern': r'(:\w+)*',
                }
            ],
            value={
                'type': str,
            },
        ),
    ])


def register():
    return EXTRA_SPEC_VALIDATORS



#    Copyright 2016 Hewlett Packard Enterprise Development Company LP
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from sqlalchemy import Index
from sqlalchemy import MetaData
from sqlalchemy import Table


def upgrade(migrate_engine):
    meta = MetaData()
    meta.bind = migrate_engine

    table = Table('console_auth_tokens', meta, autoload=True)
    for idx in table.indexes:
        if idx.columns.keys() == ['token_hash', 'instance_uuid']:
            return
    idx_name = 'console_auth_tokens_token_hash_instance_uuid_idx'
    Index(idx_name, table.c.token_hash, table.c.instance_uuid).create()



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

"""
This script concurrently builds and migrates instances. This can be useful when
troubleshooting race-conditions in virt-layer code.

Expects:

    novarc to be sourced in the environment

Helper Script for Xen Dom0:

    # cat /tmp/destroy_cache_vdis
    #!/bin/bash
    xe vdi-list | grep "Glance Image" -C1 | grep "^uuid" | awk '{print $5}' |
        xargs -n1 -I{} xe vdi-destroy uuid={}
"""
import argparse
import contextlib
import multiprocessing
import subprocess
import sys
import time

DOM0_CLEANUP_SCRIPT = "/tmp/destroy_cache_vdis"


def run(cmd):
    ret = subprocess.call(cmd, shell=True)
    if ret != 0:
        sys.stderr.write("Command exited non-zero: %s" % cmd)


@contextlib.contextmanager
def server_built(server_name, image_name, flavor=1, cleanup=True):
    run("nova boot --image=%s --flavor=%s"
        " --poll %s" % (image_name, flavor, server_name))
    try:
        yield
    finally:
        if cleanup:
            run("nova delete %s" % server_name)


@contextlib.contextmanager
def snapshot_taken(server_name, snapshot_name, cleanup=True):
    run("nova image-create %s %s"
        " --poll" % (server_name, snapshot_name))
    try:
        yield
    finally:
        if cleanup:
            run("nova image-delete %s" % snapshot_name)


def migrate_server(server_name):
    run("nova migrate %s --poll" % server_name)

    cmd = "nova list | grep %s | awk '{print $6}'" % server_name
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
    stdout, stderr = proc.communicate()
    status = stdout.strip()
    if status.upper() != 'VERIFY_RESIZE':
        sys.stderr.write("Server %s failed to rebuild" % server_name)
        return False

    # Confirm the resize
    run("nova resize-confirm %s" % server_name)
    return True


def test_migrate(context):
    count, args = context
    server_name = "server%d" % count
    cleanup = args.cleanup
    with server_built(server_name, args.image, cleanup=cleanup):
        # Migrate A -> B
        result = migrate_server(server_name)
        if not result:
            return False

        # Migrate B -> A
        return migrate_server(server_name)


def rebuild_server(server_name, snapshot_name):
    run("nova rebuild %s %s --poll" % (server_name, snapshot_name))

    cmd = "nova list | grep %s | awk '{print $6}'" % server_name
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
    stdout, stderr = proc.communicate()
    status = stdout.strip()
    if status != 'ACTIVE':
        sys.stderr.write("Server %s failed to rebuild" % server_name)
        return False

    return True


def test_rebuild(context):
    count, args = context
    server_name = "server%d" % count
    snapshot_name = "snap%d" % count
    cleanup = args.cleanup
    with server_built(server_name, args.image, cleanup=cleanup):
        with snapshot_taken(server_name, snapshot_name, cleanup=cleanup):
            return rebuild_server(server_name, snapshot_name)


def _parse_args():
    parser = argparse.ArgumentParser(
            description='Test Nova for Race Conditions.')

    parser.add_argument('tests', metavar='TESTS', type=str, nargs='*',
                        default=['rebuild', 'migrate'],
                        help='tests to run: [rebuilt|migrate]')

    parser.add_argument('-i', '--image', help="image to build from",
                        required=True)
    parser.add_argument('-n', '--num-runs', type=int, help="number of runs",
                        default=1)
    parser.add_argument('-c', '--concurrency', type=int, default=5,
                        help="number of concurrent processes")
    parser.add_argument('--no-cleanup', action='store_false', dest="cleanup",
                        default=True)
    parser.add_argument('-d', '--dom0-ips',
                        help="IP of dom0's to run cleanup script")

    return parser.parse_args()


def main():
    dom0_cleanup_script = DOM0_CLEANUP_SCRIPT
    args = _parse_args()

    if args.dom0_ips:
        dom0_ips = args.dom0_ips.split(',')
    else:
        dom0_ips = []

    start_time = time.time()
    batch_size = min(args.num_runs, args.concurrency)
    pool = multiprocessing.Pool(processes=args.concurrency)

    results = []
    for test in args.tests:
        test_func = globals().get("test_%s" % test)
        if not test_func:
            sys.stderr.write("test '%s' not found" % test)
            sys.exit(1)

        contexts = [(x, args) for x in range(args.num_runs)]

        try:
            results += pool.map(test_func, contexts)
        finally:
            if args.cleanup:
                for dom0_ip in dom0_ips:
                    run("ssh root@%s %s" % (dom0_ip, dom0_cleanup_script))

    success = all(results)
    result = "SUCCESS" if success else "FAILED"

    duration = time.time() - start_time
    print("%s, finished in %.2f secs" % (result, duration))

    sys.exit(0 if success else 1)


if __name__ == "__main__":
    main()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re

import nova.conf
from nova.virt.libvirt import utils as libvirt_utils
from nova.virt.libvirt.volume import fs
from nova.virt.libvirt.volume import remotefs

CONF = nova.conf.CONF

USERNAME_REGEX = re.compile(r"(user(?:name)?)=(?:[^ ,]+\\)?([^ ,]+)")


class LibvirtSMBFSVolumeDriver(fs.LibvirtBaseFileSystemVolumeDriver):
    """Class implements libvirt part of volume driver for SMBFS."""

    def _get_mount_point_base(self):
        return CONF.libvirt.smbfs_mount_point_base

    def get_config(self, connection_info, disk_info):
        """Returns xml for libvirt."""
        conf = super(LibvirtSMBFSVolumeDriver,
                     self).get_config(connection_info, disk_info)

        conf.source_type = 'file'
        conf.driver_cache = 'writeback'
        conf.source_path = connection_info['data']['device_path']
        conf.driver_format = connection_info['data'].get('format', 'raw')
        return conf

    def connect_volume(self, connection_info, instance):
        """Connect the volume."""
        smbfs_share = connection_info['data']['export']
        mount_path = self._get_mount_path(connection_info)

        if not libvirt_utils.is_mounted(mount_path, smbfs_share):
            mount_options = self._parse_mount_options(connection_info)
            remotefs.mount_share(mount_path, smbfs_share,
                                 export_type='cifs', options=mount_options)

        device_path = self._get_device_path(connection_info)
        connection_info['data']['device_path'] = device_path

    def disconnect_volume(self, connection_info, instance):
        """Disconnect the volume."""
        smbfs_share = connection_info['data']['export']
        mount_path = self._get_mount_path(connection_info)
        remotefs.unmount_share(mount_path, smbfs_share)

    def _parse_mount_options(self, connection_info):
        mount_options = " ".join(
            [connection_info['data'].get('options') or '',
             CONF.libvirt.smbfs_mount_options])

        if not USERNAME_REGEX.findall(mount_options):
            mount_options = mount_options + ' -o username=guest'
        else:
            # Remove the Domain Name from user name
            mount_options = USERNAME_REGEX.sub(r'\1=\2', mount_options)
        return mount_options.strip(", ").split(' ')



# Copyright 2013 NEC Corporation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from nova.api.validation import parameter_types

create = {
    'type': 'object',
    'properties': {
        'agent': {
            'type': 'object',
            'properties': {
                'hypervisor': {
                    'type': 'string', 'minLength': 0, 'maxLength': 255,
                    'pattern': '^[a-zA-Z0-9-._ ]*$'
                },
                'os': {
                    'type': 'string', 'minLength': 0, 'maxLength': 255,
                    'pattern': '^[a-zA-Z0-9-._ ]*$'
                },
                'architecture': {
                    'type': 'string', 'minLength': 0, 'maxLength': 255,
                    'pattern': '^[a-zA-Z0-9-._ ]*$'
                },
                'version': {
                    'type': 'string', 'minLength': 0, 'maxLength': 255,
                    'pattern': '^[a-zA-Z0-9-._ ]*$'
                },
                'url': {
                    'type': 'string', 'minLength': 0, 'maxLength': 255,
                    'format': 'uri'
                },
                'md5hash': {
                    'type': 'string', 'minLength': 0, 'maxLength': 255,
                    'pattern': '^[a-fA-F0-9]*$'
                },
            },
            'required': ['hypervisor', 'os', 'architecture', 'version',
                         'url', 'md5hash'],
            'additionalProperties': False,
        },
    },
    'required': ['agent'],
    'additionalProperties': False,
}


update = {
    'type': 'object',
    'properties': {
        'para': {
            'type': 'object',
            'properties': {
                'version': {
                    'type': 'string', 'minLength': 0, 'maxLength': 255,
                    'pattern': '^[a-zA-Z0-9-._ ]*$'
                },
                'url': {
                    'type': 'string', 'minLength': 0, 'maxLength': 255,
                    'format': 'uri'
                },
                'md5hash': {
                    'type': 'string', 'minLength': 0, 'maxLength': 255,
                    'pattern': '^[a-fA-F0-9]*$'
                },
            },
            'required': ['version', 'url', 'md5hash'],
            'additionalProperties': False,
        },
    },
    'required': ['para'],
    'additionalProperties': False,
}

index_query = {
    'type': 'object',
    'properties': {
        'hypervisor': parameter_types.common_query_param
    },
    # NOTE(gmann): This is kept True to keep backward compatibility.
    # As of now Schema validation stripped out the additional parameters and
    # does not raise 400. In microversion 2.75, we have blocked the additional
    # parameters.
    'additionalProperties': True
}

index_query_275 = copy.deepcopy(index_query)
index_query_275['additionalProperties'] = False



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import jsonschema
import webob

from nova.api.openstack import common
from nova.api.openstack.compute.schemas import server_tags as schema
from nova.api.openstack.compute.views import server_tags
from nova.api.openstack import wsgi
from nova.api import validation
from nova.api.validation import parameter_types
from nova.compute import api as compute
from nova.compute import vm_states
from nova import context as nova_context
from nova import exception
from nova.i18n import _
from nova.notifications import base as notifications_base
from nova import objects
from nova.policies import server_tags as st_policies


def _get_tags_names(tags):
    return [t.tag for t in tags]


def _get_instance_mapping(context, server_id):
    try:
        return objects.InstanceMapping.get_by_instance_uuid(context,
                                                            server_id)
    except exception.InstanceMappingNotFound as e:
        raise webob.exc.HTTPNotFound(explanation=e.format_message())


class ServerTagsController(wsgi.Controller):
    _view_builder_class = server_tags.ViewBuilder

    def __init__(self):
        super(ServerTagsController, self).__init__()
        self.compute_api = compute.API()

    def _check_instance_in_valid_state(self, context, server_id, action):
        instance = common.get_instance(self.compute_api, context, server_id)
        if instance.vm_state not in (vm_states.ACTIVE, vm_states.PAUSED,
                                     vm_states.SUSPENDED, vm_states.STOPPED):
            exc = exception.InstanceInvalidState(attr='vm_state',
                                                 instance_uuid=instance.uuid,
                                                 state=instance.vm_state,
                                                 method=action)
            common.raise_http_conflict_for_instance_invalid_state(exc, action,
                                                                  server_id)
        return instance

    @wsgi.Controller.api_version("2.26")
    @wsgi.response(204)
    @wsgi.expected_errors(404)
    def show(self, req, server_id, id):
        context = req.environ["nova.context"]
        im = _get_instance_mapping(context, server_id)
        context.can(st_policies.POLICY_ROOT % 'show',
                    target={'project_id': im.project_id})

        try:
            with nova_context.target_cell(context, im.cell_mapping) as cctxt:
                exists = objects.Tag.exists(cctxt, server_id, id)
        except (exception.InstanceNotFound) as e:
            raise webob.exc.HTTPNotFound(explanation=e.format_message())

        if not exists:
            msg = (_("Server %(server_id)s has no tag '%(tag)s'")
                   % {'server_id': server_id, 'tag': id})
            raise webob.exc.HTTPNotFound(explanation=msg)

    @wsgi.Controller.api_version("2.26")
    @wsgi.expected_errors(404)
    def index(self, req, server_id):
        context = req.environ["nova.context"]
        im = _get_instance_mapping(context, server_id)
        context.can(st_policies.POLICY_ROOT % 'index',
                    target={'project_id': im.project_id})

        try:
            with nova_context.target_cell(context, im.cell_mapping) as cctxt:
                tags = objects.TagList.get_by_resource_id(cctxt, server_id)
        except (exception.InstanceNotFound) as e:
            raise webob.exc.HTTPNotFound(explanation=e.format_message())

        return {'tags': _get_tags_names(tags)}

    @wsgi.Controller.api_version("2.26")
    @wsgi.expected_errors((400, 404, 409))
    @validation.schema(schema.update)
    def update(self, req, server_id, id, body):
        context = req.environ["nova.context"]
        im = _get_instance_mapping(context, server_id)
        context.can(st_policies.POLICY_ROOT % 'update',
                    target={'project_id': im.project_id})

        with nova_context.target_cell(context, im.cell_mapping) as cctxt:
            instance = self._check_instance_in_valid_state(
                cctxt, server_id, 'update tag')

        try:
            jsonschema.validate(id, parameter_types.tag)
        except jsonschema.ValidationError as e:
            msg = (_("Tag '%(tag)s' is invalid. It must be a non empty string "
                     "without characters '/' and ','. Validation error "
                     "message: %(err)s") % {'tag': id, 'err': e.message})
            raise webob.exc.HTTPBadRequest(explanation=msg)

        try:
            with nova_context.target_cell(context, im.cell_mapping) as cctxt:
                tags = objects.TagList.get_by_resource_id(cctxt, server_id)
        except exception.InstanceNotFound as e:
            raise webob.exc.HTTPNotFound(explanation=e.format_message())

        if len(tags) >= objects.instance.MAX_TAG_COUNT:
            msg = (_("The number of tags exceeded the per-server limit %d")
                   % objects.instance.MAX_TAG_COUNT)
            raise webob.exc.HTTPBadRequest(explanation=msg)

        if id in _get_tags_names(tags):
            # NOTE(snikitin): server already has specified tag
            return webob.Response(status_int=204)

        try:
            with nova_context.target_cell(context, im.cell_mapping) as cctxt:
                tag = objects.Tag(context=cctxt, resource_id=server_id, tag=id)
                tag.create()
                instance.tags = objects.TagList.get_by_resource_id(cctxt,
                                                                   server_id)
        except exception.InstanceNotFound as e:
            raise webob.exc.HTTPNotFound(explanation=e.format_message())

        notifications_base.send_instance_update_notification(
            context, instance, service="nova-api")

        response = webob.Response(status_int=201)
        response.headers['Location'] = self._view_builder.get_location(
            req, server_id, id)
        return response

    @wsgi.Controller.api_version("2.26")
    @wsgi.expected_errors((404, 409))
    @validation.schema(schema.update_all)
    def update_all(self, req, server_id, body):
        context = req.environ["nova.context"]
        im = _get_instance_mapping(context, server_id)
        context.can(st_policies.POLICY_ROOT % 'update_all',
                    target={'project_id': im.project_id})

        with nova_context.target_cell(context, im.cell_mapping) as cctxt:
            instance = self._check_instance_in_valid_state(
                cctxt, server_id, 'update tags')

        try:
            with nova_context.target_cell(context, im.cell_mapping) as cctxt:
                tags = objects.TagList.create(cctxt, server_id, body['tags'])
                instance.tags = tags
        except exception.InstanceNotFound as e:
            raise webob.exc.HTTPNotFound(explanation=e.format_message())

        notifications_base.send_instance_update_notification(
            context, instance, service="nova-api")

        return {'tags': _get_tags_names(tags)}

    @wsgi.Controller.api_version("2.26")
    @wsgi.response(204)
    @wsgi.expected_errors((404, 409))
    def delete(self, req, server_id, id):
        context = req.environ["nova.context"]
        im = _get_instance_mapping(context, server_id)
        context.can(st_policies.POLICY_ROOT % 'delete',
                    target={'project_id': im.project_id})

        with nova_context.target_cell(context, im.cell_mapping) as cctxt:
            instance = self._check_instance_in_valid_state(
                cctxt, server_id, 'delete tag')

        try:
            with nova_context.target_cell(context, im.cell_mapping) as cctxt:
                objects.Tag.destroy(cctxt, server_id, id)
                instance.tags = objects.TagList.get_by_resource_id(cctxt,
                                                                   server_id)
        except (exception.InstanceTagNotFound,
                exception.InstanceNotFound) as e:
            raise webob.exc.HTTPNotFound(explanation=e.format_message())

        notifications_base.send_instance_update_notification(
            context, instance, service="nova-api")

    @wsgi.Controller.api_version("2.26")
    @wsgi.response(204)
    @wsgi.expected_errors((404, 409))
    def delete_all(self, req, server_id):
        context = req.environ["nova.context"]
        im = _get_instance_mapping(context, server_id)
        context.can(st_policies.POLICY_ROOT % 'delete_all',
                    target={'project_id': im.project_id})

        with nova_context.target_cell(context, im.cell_mapping) as cctxt:
            instance = self._check_instance_in_valid_state(
                cctxt, server_id, 'delete tags')

        try:
            with nova_context.target_cell(context, im.cell_mapping) as cctxt:
                objects.TagList.destroy(cctxt, server_id)
                instance.tags = objects.TagList()
        except exception.InstanceNotFound as e:
            raise webob.exc.HTTPNotFound(explanation=e.format_message())

        notifications_base.send_instance_update_notification(
            context, instance, service="nova-api")
